from functools import partial

import torch
from torch import nn
from torch.nn import functional as F

from .containers import Parallel

class Residual(nn.Sequential):
    """Pass the input through a layer and add the result with the input."""
    def __init__(self, layer):
        # yapf: disable
        layers = [
            Parallel([layer, nn.Identity()]),
            Sum()
        ]
        # yapf: enable
        super().__init__(*layers)

class ModulizedFunction(nn.Module):
    """Convert a function to an nn.Module."""

    def __init__(self, fn, *args, **kwargs):
        super().__init__()
        self.fn = partial(fn, *args, **kwargs)

    def forward(self, x):
        return self.fn(x)

class Interpolate(ModulizedFunction):
    def __init__(self, mode='bilinear', align_corners=False, **kwargs):
        super().__init__(
            F.interpolate, mode='bilinear', align_corners=False, **kwargs)

class SplitTensor(nn.Module):
    """ Wrapper around `torch.split` """

    def __init__(self, size_or_sizes, dim):
        super().__init__()
        self.size_or_sizes = size_or_sizes
        self.dim = dim

    def forward(self, X):
        return X.split(self.size_or_sizes, dim=self.dim)

class Sum(nn.Module):
    def forward(self, inps):
        return sum(inps)

class AddAcross(nn.Module):
    def forward(self, inps):
        return [sum(items) for items in zip(*inps)]

class Reverse(nn.Module):
    def forward(self, inps):
        return inps[::-1]

class SelectOne(nn.Module):
    def __init__(self, idx):
        super().__init__()
        self.idx = idx

    def forward(self, xs):
        return xs[self.idx]

class Debug(nn.Module):
    def __init__(self, name):
        super().__init__()
        self.name = name

    def forward(self, x):
        print(f'{self.name}')
        print(f'type: {type(x)}, len: {len(x)}')
        print(f'shapes: {self._get_shape_recurse(x)}')
        return x

    def _get_shape_recurse(self, x):
        if isinstance(x, torch.Tensor):
            return x.shape
        return [self._get_shape_recurse(a) for a in x]


import numpy as np

def save_list(list, file_path):
    with open(file_path, 'w') as f:
        for item in list:
            f.write("%s\n" % item)

def load_list(file_path):
    with open(file_path) as f:
        list = f.read().splitlines()
    return list

def save_dict(dict, file_path):
    with open(file_path, 'w') as f:
        for key, value in dict.items():
            f.write('%s:%s\n' % (key, value))

def load_dict(file_path):
    with open(file_path) as f:
        dict = {}
        for line in f:
            key, value = line.strip().split(':')
            dict[key] = value
    return dict

def save_array(array, file_path):
    np.save(file_path, array)

def load_array(file_path):
    return np.load(file_path)


import cv2
import numpy as np
import torch
DEFAULT_CHUNK_SIZE = 1000

def get_cell_size(mask, chunk_size = DEFAULT_CHUNK_SIZE):
    """
    Get the size of each cell in the mask.

    Args:
        mask (torch.Tensor): Input segmentation mask of shape (H, W), where each unique value represents a cell.

    Returns:
        torch.Tensor: Cell sizes of shape (C,), where C is the number of cells.
    """
    unique_cells = torch.unique(mask)
    if unique_cells[0] == 0:
        unique_cells = unique_cells[1:]

    cell_sizes = torch.zeros(unique_cells.size(0), device=mask.device, dtype=mask.dtype)
    for i in range(0, unique_cells.size(0), chunk_size):
        low = i
        up = min(i + chunk_size, unique_cells.size(0))
        cell_sizes[low:up] = (mask[None, :, :] == unique_cells[low:up, None, None]).sum((1, 2))

    return cell_sizes

def get_cell_location(mask, chunk_size = DEFAULT_CHUNK_SIZE):
    unique_cells = torch.unique(mask)
    if unique_cells[0] == 0:
        unique_cells = unique_cells[1:]

    cell_locations = torch.zeros((unique_cells.size(0), 2), device=mask.device, dtype=mask.dtype)
    for i in range(0, unique_cells.size(0), chunk_size):
        low = i
        up = min(i + chunk_size, unique_cells.size(0))
        per_cell_mask = split_mask_chunk(mask, low, up)
        cell_location_x = torch.stack([torch.nonzero(per_cell_mask[i], as_tuple=True)[0].float().mean() for i in
                                       range(per_cell_mask.shape[0])])
        cell_location_y = torch.stack([torch.nonzero(per_cell_mask[i], as_tuple=True)[1].float().mean() for i in
                                       range(per_cell_mask.shape[0])])
        cell_locations[low:up] = torch.stack([cell_location_x, cell_location_y], dim=1)

    return cell_locations

def split_mask_chunk(mask, low, up):
    """
    Split the mask into individual binary masks for each cell.

    Args:
        mask (torch.Tensor): Input segmentation mask of shape (H, W), where each unique value represents a cell.
        low (int): Lower bound of the cell IDs to process.
        up (int): Upper bound of the cell IDs to process.

    Returns:
        torch.Tensor: Binary masks of shape (C, H, W), where C is the number of cells.
    """
    unique_cells = torch.unique(mask)
    if unique_cells[0] == 0:
        unique_cells = unique_cells[1:]
    binary_masks_chunk = (mask[None, :, :] == unique_cells[low:up, None, None]).float()
    return binary_masks_chunk

def aggregate_gene_expression(mask, gene_map, method='sum'):
    """
    Aggregate gene expression per cell by summing up all gene expressions inside each cell's mask.

    Args:
        mask (torch.Tensor): Input segmentation mask of shape (H, W), where each unique value represents a cell.
        gene_map (torch.Tensor): Gene expression map of shape (H, W, G), where G is the number of genes.

    Returns:
        torch.Tensor: Cell-level gene expression matrix of shape (C, G), where C is the number of cells.
    """
    # Ensure both inputs are on the GPU
    mask = mask.to('cuda') if not mask.is_cuda else mask
    gene_map = gene_map.to('cuda') if not gene_map.is_cuda else gene_map

    # Flatten spatial dimensions (H, W) to a single dimension
    H, W, G = gene_map.shape
    flattened_mask = mask.view(-1)  # Shape: (H*W,)
    flattened_genes = gene_map.view(-1, G)  # Shape: (H*W, G)

    # Get unique cell IDs and a mapping from pixels to cell indices
    unique_cells, inverse_indices = torch.unique(flattened_mask, return_inverse=True)

    # If background (0) is present, exclude it
    if unique_cells[0] == 0:
        unique_cells = unique_cells[1:]
        valid_mask = (flattened_mask != 0)
        flattened_genes = flattened_genes[valid_mask]
        inverse_indices = inverse_indices[valid_mask] - 1  # Adjust indices to remove background cell

    # Aggregate gene expressions using scatter_add
    C = unique_cells.size(0)  # Number of cells
    aggregated_expression = torch.zeros((C, G), device=gene_map.device, dtype=gene_map.dtype)
    aggregated_expression = aggregated_expression.index_add(0, inverse_indices, flattened_genes)

    if method == 'mean':
        # Count the number of pixels in each cell
        # First split the mask into individual binary masks for each cell
        # binary_masks = split_mask(mask)
        # cell_sizes = binary_masks.sum((1, 2))
        cell_sizes = get_cell_size(mask)
        # Compute the mean gene expression per cell
        aggregated_expression /= cell_sizes[:, None]

    return aggregated_expression


def sigmoid(z):
    return 1/(1 + np.exp(-z))

def get_softmask_fast(cell_nuclei_mask, tau=5):
    # Convert mask to uint8
    binary_mask = (cell_nuclei_mask > 0).astype(np.uint8)

    # Compute distance transform (distance to nearest zero pixel)
    dist_transform = cv2.distanceTransform(1 - binary_mask, cv2.DIST_L2, 5)

    # Apply sigmoid function
    soft_mask = sigmoid(dist_transform / tau)

    return 1 - soft_mask


from typing import Tuple, Optional, Type

from torch import nn

from .containers import (Parallel, SequentialMultiInputMultiOutput,
                        SequentialMultiOutput)
from .layers import (Sum, SplitTensor)

class ResNetFeatureMapsExtractor(nn.Module):
    def __init__(self, model: nn.Module, mode: Optional[str] = None):
        super().__init__()
        self.mode = mode
        # yapf: disable
        stem = nn.Sequential(
            model.conv1,
            model.bn1,
            model.relu,
            model.maxpool
        )
        layers = [
            model.layer1,
            model.layer2,
            model.layer3,
            model.layer4,
        ]
        if mode == 'fusion':
            # allow each layer to take in multiple inputs - summming them
            # before passing them through. This allows each layer to take
            # in feature maps from multiple backbones.
            multi_input_layers = [
                nn.Sequential(Sum(), layer) for layer in layers
            ]
            self.m = SequentialMultiInputMultiOutput(
                stem,
                *multi_input_layers,
                Sum()
            )
        else:
            self.m = SequentialMultiOutput(stem, *layers)

    def forward(self, x):
        if self.mode != 'fusion':
            return self.m(x)
        x, in_feats = x
        return self.m((x, *in_feats))


from typing import Any

from torch import nn

class Parallel(nn.ModuleList):
    ''' Passes inputs through multiple `nn.Module`s in parallel.
    Returns a tuple of outputs.
    '''
    def forward(self, xs: Any) -> tuple:
        # if multiple inputs, pass the 1st input through the 1st module,
        # the 2nd input through the 2nd module, etc.
        if isinstance(xs, (list, tuple)):
            return tuple(m(x) for m, x in zip(self, xs))
        # if single input, pass it through all modules
        return tuple(m(xs) for m in self)

class SequentialMultiOutput(nn.Sequential):
    """
    Like nn.Squential but returns all intermediate outputs as a tuple.

      input
        │
        │
        V
    [1st layer]───────> 1st out
        │
        │
        V
    [2nd layer]───────> 2nd out
        │
        │
        V
        .
        .
        .
        │
        │
        V
    [nth layer]───────> nth out

    """
    def forward(self, x: Any) -> tuple:
        outs = [None] * len(self)
        last_out = x
        for i, module in enumerate(self):
            last_out = module(last_out)
            outs[i] = last_out
        return tuple(outs)

class SequentialMultiInputMultiOutput(nn.Sequential):
    """
    Takes in either
    (1) an (n+1)-tuple of the form
      (last_out, 1st input, 2nd input, ..., nth input), or
    (2) an n-tuple of the form
      (1st input, 2nd input, ..., nth input),
    where n is the length of this sequential.

    If (2), the first layer in this sequential should be able to accept
    a single input. All others are expected to accept a 2-tuple of inputs.

    Returns an n-tuple of all outputs of the form:
    (1st out, 2nd out, ..., nth out).

    In other words: the ith layer in this sequential takes in as inputs the
    ith input and the output of the last layer i.e. the (i-1)th layer.
    For the 1st layer, the "output of the last layer" is last_out.

                       last_out
                      (optional)
                          │
                          │
                          V
    1st input ───────[1st layer]───────> 1st out
                          │
                          │
                          V
    2nd input ───────[2nd layer]───────> 2nd out
                          │
                          │
                          V
        .                 .                  .
        .                 .                  .
        .                 .                  .
                          │
                          │
                          V
    nth input ───────[nth layer]───────> nth out
    """
    def forward(self, xs: tuple) -> tuple:
        outs = [None] * len(self)

        if len(xs) == len(self) + 1:
            last_out = xs[0]
            layer_inputs = xs[1:]
            layers = self
            start_idx = 0
        elif len(xs) == len(self):
            last_out = self[0](xs[0])
            layer_inputs = xs[1:]
            layers = self[1:]
            outs[0] = last_out
            start_idx = 1
        else:
            raise ValueError('Invalid input format.')

        for i, (layer, x) in enumerate(zip(layers, layer_inputs), start_idx):
            last_out = layer((x, last_out))
            outs[i] = last_out

        return tuple(outs)


import numpy as np
import torch
import torch.utils.data as data
from utils.io import load_array

class SSTDataset(data.Dataset):
    def __init__(self, manager):
        self.manager = manager
        # opt, logger, log_dir
        self.opt = manager.get_opt()
        self.logger = manager.get_logger()
        self.log_dir = manager.get_log_dir()
        # Gene Map data, Nuclei Mask data
        self.logger.info(f"Loading gene map from {self.opt.gene_map}")
        # self.gene_map = tifffile.imread(self.opt.gene_map)
        self.gene_map = load_array(self.opt.gene_map)
        self.logger.info(f"Gene map shape: {self.gene_map.shape}")
        self.logger.info(f"Loading nuclei mask from {self.opt.nuclei_mask}")
        if self.opt.nuclei_mask is None:
            self.nuclei_mask = np.ones(self.gene_map.shape[:2])
        else:
            self.nuclei_mask = load_array(self.opt.nuclei_mask)
        self.nuclei_mask = self.nuclei_mask.astype(np.int32)
        self.logger.info(
            f"Nuclei mask unique values: {len(np.unique(self.nuclei_mask))}, Shape: {self.nuclei_mask.shape}")

        # Assert the shape of gene_map and nuclei_mask_fp
        assert self.gene_map.shape[:2] == self.nuclei_mask.shape, "Gene map and nuclei mask shape mismatch"
        self.patch_size = self.opt.patch_size

        # Coords list of small patch
        h_starts = list(np.arange(0, self.gene_map.shape[0] - self.patch_size, self.patch_size))
        w_starts = list(np.arange(0, self.gene_map.shape[1] - self.patch_size, self.patch_size))
        h_starts.append(self.gene_map.shape[0] - self.patch_size)
        w_starts.append(self.gene_map.shape[1] - self.patch_size)

        self.coords_starts = [(x, y) for x in h_starts for y in w_starts]
        self.logger.info(f"Total {len(self.coords_starts)} patches of size {self.patch_size}x{self.patch_size}")
        # Shifting state: horizontal/vertical/off
        self.shifting_state = "off"
        self.new_seg_mask = None

    def get_coords_index(self, coords_h1, coords_w1):
        if self.shifting_state == "horizontal":
            return self.coords_starts.index((coords_h1, coords_w1 - self.patch_size // 4))
        elif self.shifting_state == "vertical":
            return self.coords_starts.index((coords_h1 - self.patch_size // 4, coords_w1))
        else:
            return self.coords_starts.index((coords_h1, coords_w1))

    def setting_new_segmentation(self, new_seg_mask):
        self.new_seg_mask = new_seg_mask

    def __len__(self):
        return len(self.coords_starts)

    def get_gene_num(self):
        return self.gene_map.shape[-1]

    def get_nuclei_shape(self):
        return self.nuclei_mask.shape

    def set_shifting_state(self, state):
        # Ensure the state is in ["horizontal", "vertical", "off"]
        assert state in ["horizontal", "vertical", "off"], "Shifting state must be horizontal/vertical/off"
        self.shifting_state = state


    def __getitem__(self, index):
        coords = self.coords_starts[index]
        coords_h1 = coords[0]
        coords_w1 = coords[1]
        if self.shifting_state == "horizontal":
            coords_w1 = coords_w1 + self.patch_size // 4
        elif self.shifting_state == "vertical":
            coords_h1 = coords_h1 + self.patch_size // 4
        coords_h2 = coords_h1 + self.patch_size
        coords_w2 = coords_w1 + self.patch_size

        expr = self.gene_map[coords_h1:coords_h2, coords_w1:coords_w2]

        if self.new_seg_mask is not None:
            nucl = self.new_seg_mask[coords_h1:coords_h2, coords_w1:coords_w2]
        else:
            nucl = self.nuclei_mask[coords_h1:coords_h2, coords_w1:coords_w2]

        # if the size is not patch_size x patch_size due to the edge, padding zero
        # expr: H*W*C, nucl: H*W
        if expr.shape[0] < self.patch_size:
            pad_h = self.patch_size - expr.shape[0]
            expr = np.pad(expr, ((0, pad_h), (0, 0), (0, 0)), 'constant', constant_values=0)
            nucl = np.pad(nucl, ((0, pad_h), (0, 0)), 'constant', constant_values=0)
        if expr.shape[1] < self.patch_size:
            pad_w = self.patch_size - expr.shape[1]
            expr = np.pad(expr, ((0, 0), (0, pad_w), (0, 0)), 'constant', constant_values=0)
            nucl = np.pad(nucl, ((0, 0), (0, pad_w)), 'constant', constant_values=0)

        expr_torch = torch.from_numpy(expr).permute(2, 0, 1)
        nucl_torch = torch.from_numpy(nucl)

        return expr_torch, nucl_torch, coords_h1, coords_w1





from typing import Tuple, Sequence, Optional, Iterable

from torch import nn

from .containers import (Parallel, SequentialMultiInputMultiOutput)
from .layers import (Interpolate, Reverse, Sum)

class FPN(nn.Sequential):
    """
       Implementation of the architecture described in the paper
       "Feature Pyramid Networks for Object Detection" by Lin et al.,
       https://arxiv.com/abs/1612.03144.

       Takes in an n-tuple of feature maps in reverse order
       (1st feature map, 2nd feature map, ..., nth feature map), where
       the 1st feature map is the one produced by the earliest layer in the
       backbone network.

       The feature maps are passed through the architecture shown below, producing
       n outputs, such that the height and width of the ith output is equal to
       that of the corresponding input feature map and the number of channels
       is equal to out_channels.

       Returns all outputs as a tuple like so: (1st out, 2nd out, ..., nth out)

       Architecture diagram:

       nth feat. map ────────[nth in_conv]──────────┐────────[nth out_conv]────> nth out
                                                    │
                                                [upsample]
                                                    │
                                                    V
       (n-1)th feat. map ────[(n-1)th in_conv]────>(+)────[(n-1)th out_conv]────> (n-1)th out
                                                    │
                                                [upsample]
                                                    │
                                                    V
               .                     .                           .                    .
               .                     .                           .                    .
               .                     .                           .                    .
                                                    │
                                                [upsample]
                                                    │
                                                    V
       1st feat. map ────────[1st in_conv]────────>(+)────────[1st out_conv]────> 1st out

       """
    def __init__(self,
                 in_feats_shapes: Sequence[Tuple[int, ...]],
                 hidden_channels: int = 256,
                 out_channels: int = 2):
        """Constructor.
        Args:
            in_feats_shapes (Sequence[Tuple[int, ...]]): Shapes of the feature
                maps that will be fed into the network. These are expected to
                be tuples of the form (., C, H, ...).
            hidden_channels (int, optional): The number of channels to which
                all feature maps are convereted before being added together.
                Defaults to 256.
            out_channels (int, optional): Number of output channels. This will
                normally be the number of classes. Defaults to 2.
        """
        # reverse so that the deepest (i.e. produced by the deepest layer in
        # the backbone network) feature map is first.
        in_feats_shapes = in_feats_shapes[::-1]
        in_feats_channels = [s[1] for s in in_feats_shapes]
        # 1x1 conv to make the channels of all feature maps the same
        in_convs = Parallel([
            nn.Conv2d(in_channels, hidden_channels, kernel_size=1)
            for in_channels in in_feats_channels
        ])

        # yapf: disable
        def resize_and_add(to_size):
            return nn.Sequential(
                Parallel([nn.Identity(), Interpolate(size=to_size)]),
                Sum()
            )

        top_down_layer = SequentialMultiInputMultiOutput(
            nn.Identity(),
            *[resize_and_add(shape[-2:]) for shape in in_feats_shapes[1:]]
        )

        out_convs = Parallel([
            nn.Conv2d(hidden_channels, out_channels, kernel_size=3, padding=1)
            for _ in in_feats_shapes
        ])
        layers = [
            Reverse(),
            in_convs,
            top_down_layer,
            out_convs,
            Reverse()
        ]
        # yapf: enable
        super().__init__(*layers)

class PanopticFPN(nn.Sequential):
    """
    Implementation of the architecture described in the paper
    "Panoptic Feature Pyramid Networks" by Kirilov et al.,
    https://arxiv.com/abs/1901.02446.

    Takes in an n-tuple of feature maps in reverse order
    (1st feature map, 2nd feature map, ..., nth feature map), where
    the 1st feature map is the one produced by the earliest layer in the
    backbone network.

    The feature maps are passed through the architecture shown below, producing
    a single final output, with out_channels channels.

    Architecture diagram:

    nth feat. map ────[nth in_conv]─────────>───[nth upsampler]──────────┐
                                                                         │
                                                                         │
                                                                         V
    (n-1)th feat. map ──[(n-1)th in_conv]───>───[(n-1)th upsampler]────>(+)
                                                                         │
                                                                         │
                                                                         V
          .                     .                     .
          .                     .                     .
          .                     .                     .
                                                                         │
                                                                         │
                                                                         V
    1st feat. map ────[1st in_conv]─────────>───[1st upsampler]─────────(+)
                                                                         │
                                                                         │
                                                                         V
                                                                        out
    """
    def __init__(self,
                 in_feats_shapes: Sequence[Tuple[int, ...]],
                 hidden_channels: int = 256,
                 out_channels: int = 2,
                 out_size: Optional[int] = None,
                 num_upsamples_per_layer: Optional[Sequence[int]] = None,
                 upsamplng_factor: int = 2,
                 num_groups_for_norm: int = 32):
        """Constructor.

        Args:
            in_feats_shapes (Sequence[Tuple[int, ...]]): Shapes of the feature
                maps that will be fed into the network. These are expected to
                be tuples of the form (., C, H, ...).
            hidden_channels (int, optional): The number of channels to which
                all feature maps are convereted before being added together.
                Defaults to 256.
            out_channels (int, optional): Number of output channels. This will
                normally be the number of classes. Defaults to 2.
            out_size (Optional[int], optional): Size of output. If None,
                the size of the first feature map will be used.
                Defaults to None.
            num_upsamples_per_layer (Optional[Sequence[int]], optional): Number
                of upsampling iterations for each feature map. Will depend on
                the size of the feature map. Each upsampling iteration
                comprises a conv-group_norm-relu block followed by a scaling
                using torch.nn.functional.interpolate.
                If None, each feature map is assumed to be half the size of the
                preceeding one, meaning that it requires one more upsampling
                iteration than the last one.
                Defaults to None.
            upsamplng_factor (int, optional): How much to scale per upsampling
                iteration. Defaults to 2.
            num_groups_for_norm (int, optional): Number of groups for group
                norm layers. Defaults to 32.
        """
        if num_upsamples_per_layer is None:
            num_upsamples_per_layer = list(range(len(in_feats_shapes)))
        if out_size is None:
            out_size = in_feats_shapes[0][-2:]

        in_convs = Parallel([
            nn.Conv2d(s[1], hidden_channels, kernel_size=1)
            for s in in_feats_shapes
        ])

        upsamplers = self._make_upsamplers(
            in_channels=hidden_channels,
            size=out_size,
            num_upsamples_per_layer=num_upsamples_per_layer,
            num_groups=num_groups_for_norm)
        out_conv = nn.Conv2d(hidden_channels // 2, out_channels, kernel_size=1)
        # yapf: disable
        layers = [
            in_convs,
            upsamplers,
            Sum(),
            out_conv
        ]
        # yapf: enable
        super().__init__(*layers)

    @classmethod
    def _make_upsamplers(cls,
                         in_channels: int,
                         size: int,
                         num_upsamples_per_layer: Iterable[int],
                         num_groups: int = 32) -> Parallel:
        layers = []
        for num_upsamples in num_upsamples_per_layer:
            upsampler = cls._upsample_feat(
                in_channels=in_channels,
                num_upsamples=num_upsamples,
                size=size,
                num_groups=num_groups)
            layers.append(upsampler)

        upsamplers = Parallel(layers)
        return upsamplers

    @classmethod
    def _upsample_feat(cls,
                       in_channels: int,
                       num_upsamples: int,
                       size: int,
                       scale_factor: float = 2.,
                       num_groups: int = 32) -> nn.Sequential:
        if num_upsamples == 0:
            return cls._make_upsampling_block(
                in_channels=in_channels,
                out_channels=in_channels // 2,
                scale=1,
                num_groups=num_groups)
        blocks = []
        for _ in range(num_upsamples - 1):
            blocks.append(
                cls._make_upsampling_block(
                    in_channels=in_channels,
                    out_channels=in_channels,
                    scale=scale_factor,
                    num_groups=num_groups))
        blocks.append(
            cls._make_upsampling_block(
                in_channels=in_channels,
                out_channels=in_channels // 2,
                size=size,
                num_groups=num_groups))
        return nn.Sequential(*blocks)

    @classmethod
    def _make_upsampling_block(cls,
                               in_channels: int,
                               out_channels: int = None,
                               scale: float = 2,
                               size: int = None,
                               num_groups: int = 32) -> nn.Sequential:
        if out_channels is None:
            out_channels = in_channels

        # conv block that preserves size
        conv_block = [
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.GroupNorm(num_channels=out_channels, num_groups=num_groups),
            nn.ReLU(inplace=True)
        ]
        if scale == 1:
            # don't upsample
            return nn.Sequential(*conv_block)

        if size is None:
            upsample_layer = Interpolate(scale_factor=scale)
        else:
            upsample_layer = Interpolate(size=size)

        conv_block.append(upsample_layer)
        return nn.Sequential(*conv_block)






import os

import torch
from scipy.ndimage import center_of_mass
import cv2
import anndata
import distinctipy
import pandas as pd
import shutil
import wandb
import einops
from matplotlib import pyplot as plt
from torch.utils.data import DataLoader
from .fpn.factory import make_spot_fpn_resnet
from .utils.networks import *
from .utils.operate import aggregate_gene_expression, split_mask_chunk, get_cell_size, get_cell_location, get_softmask_fast
import numpy as np
from .utils.io import load_list
from torch import nn

class CellARTModel(nn.Module):
    def __init__(self, manager, gene_map_shape, patch_num):
        super(CellARTModel, self).__init__()
        self.manager = manager
        self.opt = manager.get_opt()
        self.logger = manager.get_logger()
        self.patch_num = patch_num
        # CPU Training is not supported
        self.device = torch.device("cuda")
        self.mask_size = (gene_map_shape[0], gene_map_shape[1])
        self.gene_num = gene_map_shape[2]
        self.gene_names = load_list(self.opt.gene_names)
        self.cell_chunk_size = self.opt.cell_chunk_size
        if not self.opt.nmf:
            self.celltype_names = load_list(self.opt.celltype_names)
            self.basis = np.load(self.opt.basis)
            self.basis = torch.from_numpy(self.basis).to(torch.float32).to(self.device)
            self.logger.info(f"Basis shape: {self.basis.shape}")
        else:
            # celltype_names: 1, 2, ..., factor_num
            self.celltype_names = [str(i) for i in range(self.opt.factor_num)]
            self.basis = torch.zeros(self.opt.factor_num, self.gene_num, device=self.device)
            self.logger.info(f"NMF basis shape: {self.basis.shape}")

        self._build_network()
        self._get_optimizer()

        self.new_segmentation_mask = None

    def _build_network(self, patch_emb_dim=16):
        model_fpn = make_spot_fpn_resnet(out_size=(self.opt.patch_size, self.opt.patch_size),
                                         in_channels=self.gene_num, num_classes=self.opt.deconv_emb_dim)
        self.backbone = model_fpn[0].to(self.device)
        self.fpn = model_fpn[1:].to(self.device)

        if not self.opt.nmf:
            print("=======> Cellsegmentation + CellAnnotation")
            self.deconv = DeconvNetPerSpotPatchEffect(gene_num=self.gene_num, hidden_dims=self.opt.deconv_emb_dim,
                                n_celltypes=self.basis.shape[0], patch_num=self.patch_num).to(self.device)
            if self.opt.no_patch_effect:
                print("=======> No Patch Effect")
                self.deconv = DeconvNetNoPatchEffect(gene_num=self.gene_num, hidden_dims=self.opt.deconv_emb_dim,
                                n_celltypes=self.basis.shape[0], patch_num=self.patch_num).to(self.device)
        else:
            print("=======> Cellsegmentation + NMF")
            self.deconv = NMFNet(gene_num=self.gene_num, hidden_dims=self.opt.deconv_emb_dim,
                                n_celltypes=self.basis.shape[0], patch_num=self.patch_num).to(self.device)

        self.decoder = torch.nn.Sequential(
            torch.nn.Linear(self.opt.deconv_emb_dim + patch_emb_dim, 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, 128),
            torch.nn.ReLU(),
            torch.nn.Linear(128, self.gene_num)
        ).to(self.device)
        self.slice_emb = torch.nn.Embedding(self.patch_num, patch_emb_dim).to(self.device)

    def _get_optimizer(self):
        self.optimizer_bb = torch.optim.Adam(self.backbone.parameters(), lr=self.opt.lr_bb)
        self.optimizer_fpn = torch.optim.Adam(self.fpn.parameters(), lr=self.opt.lr_fpn)
        self.optimizer_deconv = torch.optim.Adam(self.deconv.parameters(), lr=self.opt.lr_deconv)
        self.optimizer_decoder = torch.optim.Adam(self.decoder.parameters(), lr=self.opt.lr_decoder)
        self.optimizer_slice_emb = torch.optim.Adam(self.slice_emb.parameters(), lr=self.opt.lr_decoder)
        self.optimizers = [self.optimizer_bb, self.optimizer_fpn, self.optimizer_deconv, self.optimizer_decoder, self.optimizer_slice_emb]

    def _get_batch(self, batch):
        expr, nucl, coords_h1, coords_w1 = batch
        expr, nucl = (
            expr[0].to(self.device).float(), nucl[0].to(self.device).long())
        return expr, nucl, coords_h1, coords_w1

    def _filter_border(self, nucl):
        border_nucl = torch.zeros_like(nucl)
        border_nucl[:, 0] = nucl[:, 0]
        border_nucl[:, -1] = nucl[:, -1]
        border_nucl[0, :] = nucl[0, :]
        border_nucl[-1, :] = nucl[-1, :]
        # Cell id of border nuclei excluding 0
        border_nucl_id = torch.unique(border_nucl)[1:]
        # Find the index of border nuclei in nucl
        border_nucl_idx = torch.where(torch.isin(nucl, border_nucl_id))
        # Set to zero
        nucl[border_nucl_idx] = 0

        return nucl, border_nucl_id

    def _forward_feature_map(self, expr):
        # CAUTION: The input expr is not normalized, using inplace operation to avoid memory issue
        expr /= expr.sum(dim=0, keepdim=True) + 1e-6
        expr *= self.opt.normalized_total
        expr = torch.log1p(expr)
        backbone_out = self.backbone(expr.unsqueeze(0))
        feature_map = self.fpn(backbone_out)[0]
        return feature_map

    def train_model(self, train_dataset):
        train_loader = DataLoader(train_dataset, batch_size=1,
                                  shuffle=True, num_workers=self.opt.num_workers)
        deconv_loss_sum, decoder_loss_sum, total_spots = 0, 0, 0
        for epoch in range(1, self.opt.epochs + 1):
            for i, batch in enumerate(train_loader):
                self.train()
                self.backbone.apply(deactivate_batchnorm)

                expr, nucl, coords_h1, coords_w1 = self._get_batch(batch)
                patch_index = train_dataset.get_coords_index(coords_h1[0], coords_w1[0])

                # Filter border nuclei
                nucl, _ = self._filter_border(nucl)
                # Aggregate gene expression
                cell_gene = aggregate_gene_expression(nucl, einops.rearrange(expr, 'c h w -> h w c'), method='sum')
                for optimizer in self.optimizers:
                    optimizer.zero_grad()

                # Forward
                feature_map = self._forward_feature_map(expr)

                # Aggregate feature map
                cell_feature = aggregate_gene_expression(nucl, einops.rearrange(feature_map, 'c h w -> h w c'), method='mean')

                cell_umi = cell_gene.sum(dim=-1)
                cell_area = get_cell_size(nucl)
                # Filtered out small nuclei cell
                cell_feature = cell_feature[cell_area > self.opt.min_area]
                cell_gene = cell_gene[cell_area > self.opt.min_area]
                cell_umi = cell_umi[cell_area > self.opt.min_area]
                # Filtered out cell with small UMI
                cell_feature = cell_feature[cell_umi > self.opt.min_umi]
                cell_gene = cell_gene[cell_umi > self.opt.min_umi]
                cell_umi = cell_umi[cell_umi > self.opt.min_umi]

                if cell_feature.shape[0] == 0:
                    continue

                loss_deconv= self.deconv(cell_feature, cell_gene, cell_umi.unsqueeze(-1), self.basis, patch_index)
                if epoch > 1 and loss_deconv > 5e3:
                    self.logger.info(f"Epoch {epoch}, iter {i}, Skip patch {patch_index} due to some cell loss is too large: {loss_deconv}")
                    continue

                # Reconstruction
                patch_index = torch.tensor(patch_index).unsqueeze(0).to(cell_feature.device)
                patch_index = patch_index.repeat(cell_feature.shape[0])
                patch_emb = self.slice_emb(patch_index)
                z = self.decoder(torch.cat([cell_feature, patch_emb], dim=1))
                loss_decoder = torch.nn.functional.mse_loss(z, cell_gene) * self.opt.recon_loss_weight

                loss = loss_deconv + loss_decoder
                decoder_loss_sum += loss_decoder.item() * z.shape[0]
                deconv_loss_sum += loss_deconv.item() * z.shape[0]
                total_spots += z.shape[0]

                loss.backward()
                # Clip gradient
                torch.nn.utils.clip_grad_norm_(self.parameters(), self.opt.gradient_clip)

                for optimizer in self.optimizers:
                    optimizer.step()

                self.logger.info(
                    f'Epoch {epoch}/{self.opt.epochs} | iter {i} | deconv_loss {loss_deconv.item()} | decoder_loss {loss_decoder.item()} | #cell {z.shape[0]}/{cell_area.shape[0]}')
                del expr, nucl, cell_gene, cell_feature, cell_umi, cell_area, feature_map, z, loss, loss_deconv, loss_decoder
                torch.cuda.empty_cache()


            if epoch % self.opt.save_period == 0:
                self.logger.info(f"Epoch {epoch}, saving model...")
                self.save(self.manager.get_checkpoint_dir() + '/model_epoch_{}.pth'.format(epoch))

            if epoch % self.opt.pred_period == 0:
                self.logger.info(f"Epoch {epoch}, predicting...")
                self.predict(train_dataset, save_dir=self.manager.get_log_dir() + '/epoch_{}'.format(epoch))
                adata = anndata.read_h5ad(
                    f'{self.manager.get_log_dir()}/epoch_{epoch}/cell_deconv.h5ad')
                argmax_deconv_map = np.argmax(adata.obsm['deconv_beta'], axis=1)
                # Project into celltype
                celltype_names = adata.uns['celltype_names']
                ct = [celltype_names[i] for i in argmax_deconv_map]
                adata.obs['celltype'] = ct
                colors = distinctipy.get_colors(len(celltype_names) + 1, pastel_factor=0.1, rng=15)
                fig, ax = plt.subplots(1, 1, figsize=(int(adata.obs['y'].max()/adata.obs['x'].max()*10), 10))
                for i in range(len(colors)):
                    # (0,0) is on the top left corner
                    if i == 0:
                        continue
                    x = adata.obs['x'][adata.obs['celltype'] == celltype_names[i - 1]]
                    y = adata.obs['y'][adata.obs['celltype'] == celltype_names[i - 1]]
                    ax.scatter(y, x, color=colors[i], s=0.3,
                               label=celltype_names[i - 1])
                ax.invert_yaxis()
                ax.legend(bbox_to_anchor=(1.05, 1), loc='upper left')
                fig.tight_layout()
                wandb.log({"Annotation": wandb.Image(fig)}, commit=False)

            if epoch == self.opt.deconv_warmup_epochs:
                self.training_predict_segmentation(train_dataset)
                train_dataset.setting_new_segmentation(self.new_segmentation_mask)
                print("======> Finish cell segmentation")

            wandb.log({"deconv_loss": deconv_loss_sum / total_spots,
                       "decoder_loss": decoder_loss_sum / (self.opt.recon_loss_weight + 1e-6) / total_spots})
            deconv_loss_sum, decoder_loss_sum, total_spots = 0, 0, 0

    def predict(self, test_dataset, save_dir=None):
        self.backbone.apply(deactivate_batchnorm)
        self.eval()
        test_loader = DataLoader(test_dataset, batch_size=1, shuffle=False,
                                 num_workers=self.opt.num_workers)

        # Make a temp dir
        # If save_pixel_feature is True, save_dir should be provided
        assert save_dir is not None or not self.opt.save_pixel_feature, "save_dir should be provided if save_pixel_feature is True"

        if save_dir is not None:
            os.makedirs(save_dir, exist_ok=True)
            if self.opt.save_pixel_feature:
                os.makedirs(save_dir + '/pixel_feature', exist_ok=True)

        os.makedirs(self.manager.get_log_dir() + '/temp', exist_ok=True)


        with torch.no_grad():
            all_border_nucl_id = []
            for i, batch in enumerate(test_loader):
                print(f"Off Predicting {i}/{len(test_loader)}")
                expr, nucl, coords_h1, coords_w1 = self._get_batch(batch)
                nucl, border_nucl_id = self._filter_border(nucl)
                all_border_nucl_id.append(border_nucl_id)
                patch_index = test_dataset.get_coords_index(coords_h1[0], coords_w1[0])

                cell_gene = aggregate_gene_expression(nucl, einops.rearrange(expr, 'c h w -> h w c'), method='sum')
                # Forward
                feature_map = self._forward_feature_map(expr)

                if self.opt.save_pixel_feature:
                    np.save(save_dir + f'/pixel_feature/off_{coords_h1[0]}_{coords_w1[0]}_feature.npy', feature_map.cpu().numpy().transpose(1, 2, 0))

                cell_feature = aggregate_gene_expression(nucl, einops.rearrange(feature_map, 'c h w -> h w c'),
                                                         method='mean')
                if torch.sum(nucl) == 0:
                    continue

                cell_location = get_cell_location(nucl)
                cell_location = cell_location + torch.tensor([coords_h1, coords_w1]).to(cell_location.device)
                cell_ids = torch.unique(nucl)[1:]

                beta, _ = self.deconv.deconv(cell_feature, patch_index)
                # Save in temp, with prefix "off_"
                np.save(self.manager.get_log_dir() + '/temp/off_{}_cell_gene.npy'.format(i), cell_gene.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/off_{}_cell_feature.npy'.format(i), cell_feature.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/off_{}_cell_location.npy'.format(i), cell_location.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/off_{}_cell_ids.npy'.format(i), cell_ids.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/off_{}_cell_beta.npy'.format(i), beta.cpu().numpy())


                del expr, nucl, cell_gene, cell_feature, cell_location, cell_ids, beta, feature_map
                torch.cuda.empty_cache()

            all_border_nucl_id = torch.cat(all_border_nucl_id, dim=0)
            all_border_nucl_id = set(all_border_nucl_id.cpu().numpy())

            # Deal with boundary of the patches
            test_dataset.set_shifting_state("vertical")
            for i, batch in enumerate(test_loader):
                print(f"Vertical Predicting {i}/{len(test_loader)}")
                expr, nucl, coords_h1, coords_w1 = self._get_batch(batch)
                nucl, border_nucl_id = self._filter_border(nucl)
                patch_index = test_dataset.get_coords_index(coords_h1[0], coords_w1[0])

                # Only preserve cell with border nuclei
                all_border_nucl_id_torch = torch.tensor(list(all_border_nucl_id)).to(nucl.device)
                nucl = torch.where(torch.isin(nucl, all_border_nucl_id_torch), nucl, torch.zeros_like(nucl))
                # Delete the border nuclei in all_border_nucl_id
                all_border_nucl_id = all_border_nucl_id - set(torch.unique(nucl).cpu().numpy())

                cell_gene = aggregate_gene_expression(nucl, einops.rearrange(expr, 'c h w -> h w c'), method='sum')

                # Forward
                feature_map = self._forward_feature_map(expr)

                if self.opt.save_pixel_feature:
                    np.save(save_dir + f'/pixel_feature/vertical_{coords_h1[0]}_{coords_w1[0]}_feature.npy', feature_map.cpu().numpy().transpose(1, 2, 0))

                cell_feature = aggregate_gene_expression(nucl, einops.rearrange(feature_map, 'c h w -> h w c'),
                                                         method='mean')

                if torch.sum(nucl) == 0:
                    continue

                cell_location = get_cell_location(nucl)
                cell_location = cell_location + torch.tensor([coords_h1, coords_w1]).to(cell_location.device)
                cell_ids = torch.unique(nucl)[1:]

                beta, _ = self.deconv.deconv(cell_feature, patch_index)
                # Save in temp, with prefix "vertical_"
                np.save(self.manager.get_log_dir() + '/temp/vertical_{}_cell_gene.npy'.format(i), cell_gene.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/vertical_{}_cell_feature.npy'.format(i), cell_feature.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/vertical_{}_cell_location.npy'.format(i), cell_location.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/vertical_{}_cell_ids.npy'.format(i), cell_ids.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/vertical_{}_cell_beta.npy'.format(i), beta.cpu().numpy())


                del expr, nucl, cell_gene, cell_feature, cell_location, cell_ids, beta, feature_map
                torch.cuda.empty_cache()

            test_dataset.set_shifting_state("horizontal")
            for i, batch in enumerate(test_loader):
                print(f"Horizontal Predicting {i}/{len(test_loader)}")
                expr, nucl, coords_h1, coords_w1 = self._get_batch(batch)
                nucl, border_nucl_id = self._filter_border(nucl)
                patch_index = test_dataset.get_coords_index(coords_h1[0], coords_w1[0])

                # Only preserve cell with border nuclei
                all_border_nucl_id_torch = torch.tensor(list(all_border_nucl_id)).to(nucl.device)
                nucl = torch.where(torch.isin(nucl, all_border_nucl_id_torch), nucl, torch.zeros_like(nucl))
                all_border_nucl_id = all_border_nucl_id - set(torch.unique(nucl).cpu().numpy())

                cell_gene = aggregate_gene_expression(nucl, einops.rearrange(expr, 'c h w -> h w c'), method='sum')

                # Forward
                feature_map = self._forward_feature_map(expr)

                if self.opt.save_pixel_feature:
                    np.save(save_dir + f'/pixel_feature/horizontal_{coords_h1[0]}_{coords_w1[0]}_feature.npy',
                        feature_map.cpu().numpy().transpose(1, 2, 0))

                cell_feature = aggregate_gene_expression(nucl, einops.rearrange(feature_map, 'c h w -> h w c'),
                                                         method='mean')
                if torch.sum(nucl) == 0:
                    continue

                cell_location = get_cell_location(nucl)
                cell_location = cell_location + torch.tensor([coords_h1, coords_w1]).to(cell_location.device)
                cell_ids = torch.unique(nucl)[1:]

                beta, _ = self.deconv.deconv(cell_feature, patch_index)
                # Save in temp, with prefix "horizontal_"
                np.save(self.manager.get_log_dir() + '/temp/horizontal_{}_cell_gene.npy'.format(i), cell_gene.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/horizontal_{}_cell_feature.npy'.format(i), cell_feature.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/horizontal_{}_cell_location.npy'.format(i), cell_location.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/horizontal_{}_cell_ids.npy'.format(i), cell_ids.cpu().numpy())
                np.save(self.manager.get_log_dir() + '/temp/horizontal_{}_cell_beta.npy'.format(i), beta.cpu().numpy())

                del expr, nucl, cell_gene, cell_feature, cell_location, cell_ids, beta, feature_map
                torch.cuda.empty_cache()

            test_dataset.set_shifting_state("off")

        # Concat all in the temp dir
        for prefix in ["off_", "vertical_", "horizontal_"]:
            cell_gene_list = []
            cell_feature_list = []
            cell_location_list = []
            cell_ids_list = []
            cell_beta_list = []
            for i in range(len(test_loader)):
                if os.path.exists(self.manager.get_log_dir() + f'/temp/{prefix}{i}_cell_gene.npy'):
                    cell_gene_list.append(np.load(self.manager.get_log_dir() + f'/temp/{prefix}{i}_cell_gene.npy'))
                    cell_feature_list.append(np.load(self.manager.get_log_dir() + f'/temp/{prefix}{i}_cell_feature.npy'))
                    cell_location_list.append(np.load(self.manager.get_log_dir() + f'/temp/{prefix}{i}_cell_location.npy'))
                    cell_ids_list.append(np.load(self.manager.get_log_dir() + f'/temp/{prefix}{i}_cell_ids.npy'))
                    cell_beta_list.append(np.load(self.manager.get_log_dir() + f'/temp/{prefix}{i}_cell_beta.npy'))

            cell_gene = np.concatenate(cell_gene_list, axis=0)
            cell_feature = np.concatenate(cell_feature_list, axis=0)
            cell_location = np.concatenate(cell_location_list, axis=0)
            cell_ids = np.concatenate(cell_ids_list, axis=0)
            cell_beta = np.concatenate(cell_beta_list, axis=0)

            if prefix == "off_":
                cell_gene_off = cell_gene
                cell_feature_off = cell_feature
                cell_location_off = cell_location
                cell_ids_off = cell_ids
                cell_beta_off = cell_beta
            elif prefix == "vertical_":
                cell_gene_vertical = cell_gene
                cell_feature_vertical = cell_feature
                cell_location_vertical = cell_location
                cell_ids_vertical = cell_ids
                cell_beta_vertical = cell_beta
            else:
                cell_gene_horizontal = cell_gene
                cell_feature_horizontal = cell_feature
                cell_location_horizontal = cell_location
                cell_ids_horizontal = cell_ids
                cell_beta_horizontal = cell_beta

        # Concatenate all
        cell_gene = np.concatenate([cell_gene_off, cell_gene_vertical, cell_gene_horizontal], axis=0)
        cell_feature = np.concatenate([cell_feature_off, cell_feature_vertical, cell_feature_horizontal], axis=0)
        cell_location = np.concatenate([cell_location_off, cell_location_vertical, cell_location_horizontal], axis=0)
        cell_ids = np.concatenate([cell_ids_off, cell_ids_vertical, cell_ids_horizontal], axis=0)
        cell_beta = np.concatenate([cell_beta_off, cell_beta_vertical, cell_beta_horizontal], axis=0)

        gene_names = self.gene_names
        if save_dir is None:
            save_dir = self.manager.get_log_dir()
        else:
            os.makedirs(save_dir, exist_ok=True)

        # Save Anndata
        # Give each spot a unique id
        df_obs = pd.DataFrame(index=cell_ids)
        df_obs['x'] = cell_location[:, 0]
        df_obs['y'] = cell_location[:, 1]

        adata = anndata.AnnData(X = cell_gene, obs = df_obs, var = pd.DataFrame(index = gene_names))
        adata.obsm['latent'] = cell_feature
        adata.uns['celltype_names'] = self.celltype_names
        adata.obsm['deconv_beta'] = cell_beta
        adata.obs["cell_id"] = adata.obs_names.astype(float).astype(int).astype(str)
        adata.obs.set_index("cell_id", inplace=True)
        # Since there is some overlap of the bundary patch, we need to remove the duplicated cell id
        # (last patch: shape[1]-patch_size, but last second patch is from shape[1]-patch_size to shape[1])
        adata = adata[~adata.obs.index.duplicated(keep='first')]
        temp = pd.DataFrame(adata.obsm["deconv_beta"], columns=adata.uns["celltype_names"],
                            index=adata.obs.index)
        adata.obs["celltype"] = temp.idxmax(axis=1)

        # if nmf: save basis
        if self.opt.nmf:
            adata.uns['basis'] = self.deconv.get_basis().cpu().numpy()
        else:
            adata.uns['basis'] = self.basis.cpu().numpy()

        adata.write_h5ad(save_dir + '/cell_deconv.h5ad')

        # Collect feature map
        if self.opt.save_pixel_feature:
            print("======> Collecting feature map")
            whole_feature_map = np.zeros((self.mask_size[0], self.mask_size[1], self.opt.deconv_emb_dim))
            feature_map_dir = save_dir + '/pixel_feature'
            # off_{coords_h1[0]}_{coords_w1[0]}_feature.npy
            for f in os.listdir(feature_map_dir):
                if f.endswith('.npy') and f.startswith('off_'):
                    feature_map = np.load(os.path.join(feature_map_dir, f))
                    coords_h1, coords_w1 = map(int, f.split('_')[1:3])
                    whole_feature_map[coords_h1:coords_h1 + self.opt.patch_size,
                                     coords_w1:coords_w1 + self.opt.patch_size] = feature_map

            for f in os.listdir(feature_map_dir):
                if f.endswith('.npy') and f.startswith('vertical_'):
                    feature_map = np.load(os.path.join(feature_map_dir, f))
                    coords_h1, coords_w1 = map(int, f.split('_')[1:3])
                    if coords_h1 + self.opt.patch_size > self.mask_size[0] or coords_w1 + self.opt.patch_size > self.mask_size[1]:
                        continue
                    low_h1 = coords_h1 + self.opt.patch_size - self.opt.patch_size // 4 - self.opt.patch_size // 8
                    high_h1 = coords_h1 + self.opt.patch_size - self.opt.patch_size // 4 + self.opt.patch_size // 8
                    whole_feature_map[low_h1:high_h1, coords_w1:coords_w1 + self.opt.patch_size] = feature_map[
                        (self.opt.patch_size - self.opt.patch_size // 4 - self.opt.patch_size // 8):(self.opt.patch_size - self.opt.patch_size // 4 + self.opt.patch_size // 8), :]

            for f in os.listdir(feature_map_dir):
                if f.endswith('.npy') and f.startswith('horizontal_'):
                    feature_map = np.load(os.path.join(feature_map_dir, f))
                    coords_h1, coords_w1 = map(int, f.split('_')[1:3])
                    if coords_h1 + self.opt.patch_size > self.mask_size[0] or coords_w1 + self.opt.patch_size > self.mask_size[1]:
                        continue
                    low_w1 = coords_w1 + self.opt.patch_size - self.opt.patch_size // 4 - self.opt.patch_size // 8
                    high_w1 = coords_w1 + self.opt.patch_size - self.opt.patch_size // 4 + self.opt.patch_size // 8
                    whole_feature_map[coords_h1:coords_h1 + self.opt.patch_size, low_w1:high_w1] = feature_map[:,
                        (self.opt.patch_size - self.opt.patch_size // 4 - self.opt.patch_size // 8):(self.opt.patch_size - self.opt.patch_size // 4 + self.opt.patch_size // 8)]

            # Save feature map
            np.save(save_dir + '/feature_map.npy', whole_feature_map)

            # Delete feature map dir
            shutil.rmtree(feature_map_dir)

        # Delete temp dir
        shutil.rmtree(self.manager.get_log_dir() + '/temp')

    def training_predict_segmentation(self, dataset):
        cell_chunk_size = self.cell_chunk_size
        self.eval()
        self.backbone.apply(deactivate_batchnorm)
        train_loader = DataLoader(dataset, batch_size=1, shuffle=False,
                                 num_workers=self.opt.num_workers)
        cell_predictor = CellPredictNet().cuda()
        optimizer = torch.optim.Adam(cell_predictor.parameters(), lr=1e-3,
                                     betas=(0.9,
                                            0.999),
                                     weight_decay=0.0001)
        for e in range(self.opt.seg_training_epochs):
            for ite, batch in enumerate(train_loader):
                expr, nucl, coords_h1, coords_w1 = self._get_batch(batch)
                nucl, border_nucl_id = self._filter_border(nucl)
                # Forward
                with torch.no_grad():
                    feature_map = self._forward_feature_map(expr)
                cell_ids = torch.unique(nucl)[1:]
                if len(cell_ids) == 0:
                    continue
                cell_chunks = []
                for i in range(0, len(cell_ids), cell_chunk_size):
                    if min(i + cell_chunk_size, len(cell_ids)) - i < cell_chunk_size // 2:
                        continue

                    cell_chunks.append((i, min(i + cell_chunk_size, len(cell_ids))))
                if len(cell_chunks) == 0:
                    continue

                minipatch_size = 48
                half_size = minipatch_size // 2

                feature_map_pad = torch.nn.functional.pad(feature_map, (half_size, half_size, half_size, half_size),
                                                          mode='constant', value=0).detach()
                for cell_chunk in cell_chunks:
                    binary_masks_chunk = split_mask_chunk(nucl, cell_chunk[0], cell_chunk[1])
                    N, H, W = binary_masks_chunk.shape
                    centers = []
                    binary_masks_pad = torch.nn.functional.pad(binary_masks_chunk,
                                                               (half_size, half_size, half_size, half_size),
                                                               mode='constant',
                                                               value=0)
                    for i in range(N):
                        cy, cx = center_of_mass(binary_masks_pad[i].cpu().numpy())  # Returns float coordinates
                        centers.append([int(cy), int(cx)])
                    centers = torch.tensor(centers, device=feature_map_pad.device)
                    mini_feat = torch.zeros((N, feature_map_pad.shape[0], minipatch_size, minipatch_size),
                                            device=feature_map_pad.device)
                    mini_nucl = torch.zeros((N, minipatch_size, minipatch_size), device=feature_map_pad.device)
                    for i in range(N):
                        cy, cx = centers[i]
                        cy = int(cy)
                        cx = int(cx)
                        mini_feat[i] = feature_map_pad[:, cy - half_size:cy + half_size, cx - half_size:cx + half_size]
                        mini_nucl[i] = binary_masks_pad[i, cy - half_size:cy + half_size, cx - half_size:cx + half_size]
                    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (self.opt.dilation_kernel_size, self.opt.dilation_kernel_size))
                    soft_masks = np.zeros((len(mini_nucl), mini_nucl[0].shape[0], mini_nucl[0].shape[1]))
                    dilated_masks = np.zeros((len(mini_nucl), mini_nucl[0].shape[0], mini_nucl[0].shape[1]))
                    for c in range(len(mini_nucl)):
                        cell_nuclei_mask = mini_nucl[c]
                        cell_nuclei_mask = cell_nuclei_mask.cpu().numpy()
                        I = get_softmask_fast(cell_nuclei_mask, tau=self.opt.tau)
                        soft_masks[c] = I
                        dilated_masks[c] = cv2.dilate(cell_nuclei_mask.astype('uint8'), kernel, iterations=self.opt.dilation_iter_num)

                    pred = cell_predictor(torch.from_numpy(soft_masks).float().cuda().unsqueeze(1), mini_feat)
                    criterion_ce = torch.nn.CrossEntropyLoss(reduction='none', weight=torch.tensor([1., 1.]).cuda())
                    loss_nucl = criterion_ce(pred, mini_nucl.long())
                    loss_mask = mini_nucl + torch.from_numpy(1 - dilated_masks).int().cuda()
                    loss_nucl = loss_nucl * loss_mask

                    loss = torch.sum(loss_nucl) / torch.sum(loss_mask)
                    optimizer.zero_grad()
                    loss.backward()
                    optimizer.step()
                    self.logger.info(
                        f"Epoch: {e}, Iter: {ite}, Cell chunk: {cell_chunk}, Loss_nucl: {torch.sum(loss_nucl) / torch.sum(loss_mask)}")

        self.logger.info("=======> Training done, Start predicting new cell segmentation mask")
        all_border_nucl_id = []
        new_segmentation_mask_dict = {}
        for ite, batch in enumerate(train_loader):
            print(f"Iter {ite}/{len(train_loader)} in no shift")
            expr, nucl, coords_h1, coords_w1 = self._get_batch(batch)
            nucl, border_nucl_id = self._filter_border(nucl)
            all_border_nucl_id.append(border_nucl_id)
            # Forward
            with torch.no_grad():
                feature_map = self._forward_feature_map(expr)

            cell_ids = torch.unique(nucl)[1:]
            if len(cell_ids) == 0:
                continue
            cell_chunks = []

            for i in range(0, len(cell_ids), cell_chunk_size):
                cell_chunks.append((i, min(i + cell_chunk_size, len(cell_ids))))

            minipatch_size = 48
            half_size = minipatch_size // 2
            feature_map_pad = torch.nn.functional.pad(feature_map, (half_size, half_size, half_size, half_size),
                                                      mode='constant', value=0).detach()
            all_center = []
            all_pred = []

            for cell_chunk in cell_chunks:
                binary_masks_chunk = split_mask_chunk(nucl, cell_chunk[0], cell_chunk[1])
                N, H, W = binary_masks_chunk.shape
                centers = []
                binary_masks_pad = torch.nn.functional.pad(binary_masks_chunk,
                                                           (half_size, half_size, half_size, half_size),
                                                           mode='constant', value=0)
                for i in range(N):
                    cy, cx = center_of_mass(binary_masks_pad[i].cpu().numpy())  # Returns float coordinates
                    centers.append([int(cy), int(cx)])
                centers = torch.tensor(centers, device=feature_map_pad.device)
                mini_feat = torch.zeros((N, feature_map_pad.shape[0], minipatch_size, minipatch_size),
                                        device=feature_map_pad.device)
                mini_nucl = torch.zeros((N, minipatch_size, minipatch_size), device=feature_map_pad.device)
                for i in range(N):
                    cy, cx = centers[i]
                    cy = int(cy)
                    cx = int(cx)
                    mini_feat[i] = feature_map_pad[:, cy - half_size:cy + half_size, cx - half_size:cx + half_size]
                    mini_nucl[i] = binary_masks_pad[i, cy - half_size:cy + half_size, cx - half_size:cx + half_size]
                soft_masks = np.zeros((len(mini_nucl), mini_nucl[0].shape[0], mini_nucl[0].shape[1]))
                for c in range(len(mini_nucl)):
                    cell_nuclei_mask = mini_nucl[c]
                    cell_nuclei_mask = cell_nuclei_mask.cpu().numpy()
                    I = get_softmask_fast(cell_nuclei_mask)
                    soft_masks[c] = I
                pred = cell_predictor(torch.from_numpy(soft_masks).float().cuda().unsqueeze(1), mini_feat)
                pred_prob = torch.nn.functional.softmax(pred, dim=1)[:, 1]

                all_center.append(centers.cpu().numpy())
                all_pred.append(pred_prob.detach().cpu().numpy())

            # Concatenate all the chunks
            all_center = np.concatenate(all_center)
            all_pred = np.concatenate(all_pred)
            pred_probs_pad = np.zeros((len(all_center), H + 2 * half_size, W + 2 * half_size))
            for i in range(len(all_center)):
                cy, cx = all_center[i]
                cy = int(cy)
                cx = int(cx)
                pred_probs_pad[i, cy - half_size:cy + half_size, cx - half_size:cx + half_size] = all_pred[i]
            pred_probs = pred_probs_pad[:, half_size:-half_size, half_size:-half_size]
            bgd_probs = 1 - pred_probs.max(axis=0)
            # Concat cell and background
            cell_probs = np.concatenate([bgd_probs[np.newaxis], pred_probs], axis=0)
            final_seg = np.argmax(cell_probs, axis=0)
            # Make prediciton projection to the original nucl index
            dictionary = dict(
                zip(np.arange(0, len(cell_ids) + 1), torch.cat([torch.tensor([0]), cell_ids.cpu()]).numpy()))

            final_seg = np.vectorize(dictionary.get)(final_seg)
            final_seg = np.where(nucl.cpu().numpy() > 0, nucl.cpu().numpy(), final_seg)
            # If the predicted cell touch the border, set to 0 and add to the border_nucl_id
            final_seg, border_nucl_id = self._filter_border(torch.tensor(final_seg).to(nucl.device))
            all_border_nucl_id.append(border_nucl_id)
            new_segmentation_mask_dict[(coords_h1, coords_w1)] = final_seg.cpu().numpy()

            # Delete
            del expr, nucl, feature_map

        all_border_nucl_id = torch.cat(all_border_nucl_id, dim=0)
        all_border_nucl_id = set(all_border_nucl_id.cpu().numpy())
        dataset.set_shifting_state("vertical")
        vertical_segmentation_mask_dict = {}
        for ite, batch in enumerate(train_loader):
            print(f"Iter {ite}/{len(train_loader)} in vertical shift")
            expr, nucl, coords_h1, coords_w1 = self._get_batch(batch)
            nucl, border_nucl_id = self._filter_border(nucl)
            # Only preserve cell with border nuclei
            all_border_nucl_id_torch = torch.tensor(list(all_border_nucl_id)).to(nucl.device)
            nucl = torch.where(torch.isin(nucl, all_border_nucl_id_torch), nucl, torch.zeros_like(nucl))
            all_border_nucl_id = all_border_nucl_id - set(torch.unique(nucl).cpu().numpy())
            # Forward
            with torch.no_grad():
                feature_map = self._forward_feature_map(expr)

            cell_ids = torch.unique(nucl)[1:]
            if len(cell_ids) == 0:
                continue
            # Divide the cell into chunks
            cell_chunks = []
            for i in range(0, len(cell_ids), cell_chunk_size):
                cell_chunks.append((i, min(i + cell_chunk_size, len(cell_ids))))

            minipatch_size = 48
            half_size = minipatch_size // 2

            feature_map_pad = torch.nn.functional.pad(feature_map, (half_size, half_size, half_size, half_size),
                                                        mode='constant', value=0).detach()
            all_center = []
            all_pred = []
            for cell_chunk in cell_chunks:
                binary_masks_chunk = split_mask_chunk(nucl, cell_chunk[0], cell_chunk[1])
                N, H, W = binary_masks_chunk[cell_chunk[0]:cell_chunk[1]].shape
                centers = []
                binary_masks_pad = torch.nn.functional.pad(binary_masks_chunk[cell_chunk[0]:cell_chunk[1]],
                                                           (half_size, half_size, half_size, half_size),
                                                           mode='constant', value=0)
                for i in range(N):
                    cy, cx = center_of_mass(binary_masks_pad[i].cpu().numpy())  # Returns float coordinates
                    centers.append([int(cy), int(cx)])
                centers = torch.tensor(centers, device=feature_map_pad.device)
                mini_feat = torch.zeros((N, feature_map_pad.shape[0], minipatch_size, minipatch_size),
                                        device=feature_map_pad.device)
                mini_nucl = torch.zeros((N, minipatch_size, minipatch_size), device=feature_map_pad.device)
                for i in range(N):
                    cy, cx = centers[i]
                    cy = int(cy)
                    cx = int(cx)
                    mini_feat[i] = feature_map_pad[:, cy - half_size:cy + half_size, cx - half_size:cx + half_size]
                    mini_nucl[i] = binary_masks_pad[i, cy - half_size:cy + half_size, cx - half_size:cx + half_size]
                # print(f"N: {N}, H: {H}, W: {W}, binary_masks_chunk: {binary_masks_chunk.shape}, mini_nucl: {mini_nucl.shape}")
                soft_masks = np.zeros((len(mini_nucl), mini_nucl[0].shape[0], mini_nucl[0].shape[1]))
                for c in range(len(mini_nucl)):
                    cell_nuclei_mask = mini_nucl[c]
                    cell_nuclei_mask = cell_nuclei_mask.cpu().numpy()
                    I = get_softmask_fast(cell_nuclei_mask)
                    soft_masks[c] = I

                pred = cell_predictor(torch.from_numpy(soft_masks).float().cuda().unsqueeze(1), mini_feat)
                pred_prob = torch.nn.functional.softmax(pred, dim=1)[:, 1]

                all_center.append(centers.cpu().numpy())
                all_pred.append(pred_prob.detach().cpu().numpy())
            # Concatenate all the chunks
            all_center = np.concatenate(all_center)
            all_pred = np.concatenate(all_pred)
            pred_probs_pad = np.zeros((len(all_center), H + 2 * half_size, W + 2 * half_size))
            for i in range(len(all_center)):
                cy, cx = all_center[i]
                cy = int(cy)
                cx = int(cx)
                pred_probs_pad[i, cy - half_size:cy + half_size, cx - half_size:cx + half_size] = all_pred[i]
            pred_probs = pred_probs_pad[:, half_size:-half_size, half_size:-half_size]
            bgd_probs = 1 - pred_probs.max(axis=0)
            # Concat cell and background
            cell_probs = np.concatenate([bgd_probs[np.newaxis], pred_probs], axis=0)
            final_seg = np.argmax(cell_probs, axis=0)
            # Make prediciton projection to the original nucl index
            dictionary = dict(
                zip(np.arange(0, len(cell_ids) + 1), torch.cat([torch.tensor([0]), cell_ids.cpu()]).numpy()))

            final_seg = np.vectorize(dictionary.get)(final_seg)
            final_seg = np.where(nucl.cpu().numpy() > 0, nucl.cpu().numpy(), final_seg)
            # If the predicted cell touch the border, set to 0 and add to the border_nucl_id
            final_seg, border_nucl_id = self._filter_border(torch.tensor(final_seg).to(nucl.device))
            all_border_nucl_id = all_border_nucl_id.union(set(border_nucl_id))
            vertical_segmentation_mask_dict[(coords_h1, coords_w1)] = final_seg.cpu().numpy()

            del expr, nucl, feature_map

        dataset.set_shifting_state("horizontal")
        horizontal_segmentation_mask_dict = {}
        for ite, batch in enumerate(train_loader):
            print(f"Iter {ite}/{len(train_loader)} in horizontal shift")
            expr, nucl, coords_h1, coords_w1 = self._get_batch(batch)
            nucl, border_nucl_id = self._filter_border(nucl)

            # Only preserve cell with border nuclei
            all_border_nucl_id_torch = torch.tensor(list(all_border_nucl_id)).to(nucl.device)
            nucl = torch.where(torch.isin(nucl, all_border_nucl_id_torch), nucl, torch.zeros_like(nucl))
            all_border_nucl_id = all_border_nucl_id - set(torch.unique(nucl).cpu().numpy())

            # Forward
            with torch.no_grad():
                feature_map = self._forward_feature_map(expr)
            cell_ids = torch.unique(nucl)[1:]
            if len(cell_ids) == 0:
                continue
            cell_chunks = []

            for i in range(0, len(cell_ids), cell_chunk_size):
                cell_chunks.append((i, min(i + cell_chunk_size, len(cell_ids))))

            minipatch_size = 48
            half_size = minipatch_size // 2
            feature_map_pad = torch.nn.functional.pad(feature_map, (half_size, half_size, half_size, half_size),
                                                      mode='constant', value=0).detach()

            all_center = []
            all_pred = []

            for cell_chunk in cell_chunks:
                binary_masks_chunk = split_mask_chunk(nucl, cell_chunk[0], cell_chunk[1])
                N, H, W = binary_masks_chunk.shape
                centers = []
                binary_masks_pad = torch.nn.functional.pad(binary_masks_chunk,
                                                           (half_size, half_size, half_size, half_size),
                                                           mode='constant', value=0)
                for i in range(N):
                    cy, cx = center_of_mass(binary_masks_pad[i].cpu().numpy())  # Returns float coordinates
                    centers.append([int(cy), int(cx)])
                centers = torch.tensor(centers, device=feature_map_pad.device)
                mini_feat = torch.zeros((N, feature_map_pad.shape[0], minipatch_size, minipatch_size),
                                        device=feature_map_pad.device)
                mini_nucl = torch.zeros((N, minipatch_size, minipatch_size), device=feature_map_pad.device)
                for i in range(N):
                    cy, cx = centers[i]
                    cy = int(cy)
                    cx = int(cx)
                    mini_feat[i] = feature_map_pad[:, cy - half_size:cy + half_size, cx - half_size:cx + half_size]
                    mini_nucl[i] = binary_masks_pad[i, cy - half_size:cy + half_size, cx - half_size:cx + half_size]
                print(
                    f"N: {N}, H: {H}, W: {W}, binary_masks_chunk: {binary_masks_chunk.shape}, mini_nucl: {mini_nucl.shape}")
                soft_masks = np.zeros((len(mini_nucl), mini_nucl[0].shape[0], mini_nucl[0].shape[1]))
                for c in range(len(mini_nucl)):
                    cell_nuclei_mask = mini_nucl[c]
                    cell_nuclei_mask = cell_nuclei_mask.cpu().numpy()
                    I = get_softmask_fast(cell_nuclei_mask)
                    soft_masks[c] = I

                pred = cell_predictor(torch.from_numpy(soft_masks).float().cuda().unsqueeze(1), mini_feat)
                pred_prob = torch.nn.functional.softmax(pred, dim=1)[:, 1]

                all_center.append(centers.cpu().numpy())
                all_pred.append(pred_prob.detach().cpu().numpy())
            # Concatenate all the chunks
            all_center = np.concatenate(all_center)
            all_pred = np.concatenate(all_pred)
            pred_probs_pad = np.zeros((len(all_center), H + 2 * half_size, W + 2 * half_size))
            for i in range(len(all_center)):
                cy, cx = all_center[i]
                cy = int(cy)
                cx = int(cx)
                pred_probs_pad[i, cy - half_size:cy + half_size, cx - half_size:cx + half_size] = all_pred[i]

            pred_probs = pred_probs_pad[:, half_size:-half_size, half_size:-half_size]
            bgd_probs = 1 - pred_probs.max(axis=0)
            # Concat cell and background
            cell_probs = np.concatenate([bgd_probs[np.newaxis], pred_probs], axis=0)
            final_seg = np.argmax(cell_probs, axis=0)
            # Make prediciton projection to the original nucl index
            dictionary = dict(
                zip(np.arange(0, len(cell_ids) + 1), torch.cat([torch.tensor([0]), cell_ids.cpu()]).numpy()))
            final_seg = np.vectorize(dictionary.get)(final_seg)
            final_seg = np.where(nucl.cpu().numpy() > 0, nucl.cpu().numpy(), final_seg)
            horizontal_segmentation_mask_dict[(coords_h1, coords_w1)] = final_seg

            del expr, nucl, feature_map

        dataset.set_shifting_state("off")

        # Merge all the segmentation mask
        new_segmentation_mask = np.zeros(dataset.nuclei_mask.shape)

        for coords_h1, coords_w1 in vertical_segmentation_mask_dict.keys():
            # Only update the non-zero region
            current_region = new_segmentation_mask[coords_h1:coords_h1 + self.opt.patch_size,
                             coords_w1:coords_w1 + self.opt.patch_size]
            vertical_region = vertical_segmentation_mask_dict[(coords_h1, coords_w1)][0:current_region.shape[0],
                              0:current_region.shape[1]]
            x, y = np.where(vertical_region > 0)
            current_region[x, y] = vertical_region[x, y]
            new_segmentation_mask[coords_h1:coords_h1 + self.opt.patch_size,
            coords_w1:coords_w1 + self.opt.patch_size] = current_region

        for coords_h1, coords_w1 in horizontal_segmentation_mask_dict.keys():
            # Only update the non-zero region
            current_region = new_segmentation_mask[coords_h1:coords_h1 + self.opt.patch_size,
                             coords_w1:coords_w1 + self.opt.patch_size]
            horizontal_region = horizontal_segmentation_mask_dict[(coords_h1, coords_w1)][0:current_region.shape[0],
                                0:current_region.shape[1]]
            x, y = np.where(horizontal_region > 0)
            current_region[x, y] = horizontal_region[x, y]
            new_segmentation_mask[coords_h1:coords_h1 + self.opt.patch_size,
            coords_w1:coords_w1 + self.opt.patch_size] = current_region

        for coords_h1, coords_w1 in new_segmentation_mask_dict.keys():
            current_region = new_segmentation_mask[coords_h1:coords_h1 + self.opt.patch_size,
                             coords_w1:coords_w1 + self.opt.patch_size]
            region = new_segmentation_mask_dict[(coords_h1, coords_w1)]
            x, y = np.where(region > 0)
            current_region[x, y] = region[x, y]
            new_segmentation_mask[coords_h1:coords_h1 + self.opt.patch_size,
            coords_w1:coords_w1 + self.opt.patch_size] = current_region

        self.new_segmentation_mask = new_segmentation_mask
        # Save
        np.save(self.manager.get_log_dir() + '/new_segmentation_mask.npy', new_segmentation_mask)
        # wandb show the segmentation result
        fig, ax = plt.subplots(1, 1, figsize=(10, 10))
        ax.imshow(new_segmentation_mask > 0)
        wandb.log({"new_segmentation": wandb.Image(fig)}, commit=False)

    def save(self, path):
        """Save the model state
        """
        save_dict = {'model_state': self.state_dict()}
        torch.save(save_dict, path)

    def load(self, path, strict = True):
        """Load a model state from a checkpoint file
        """
        checkpoint_file = path
        checkpoints = torch.load(checkpoint_file)
        self.load_state_dict(checkpoints['model_state'], strict=strict)




























from typing import Tuple, List

import torch
from torch import nn

def _get_shapes(model: nn.Module,
                channels: int = 3,
                size: Tuple[int, int] = (224, 224)) -> List[Tuple[int, ...]]:
    """Extract shapes of feature maps computed by the model.

    The model must be an nn.Module whose __call__ method returns all feature
    maps when called with an input.
    """
    # save state so we can restore laterD
    state = model.training

    model.eval()
    with torch.no_grad():
        x = torch.empty(1, channels, *size)
        feats = model(x)

    # restore state
    model.train(state)

    if isinstance(feats, torch.Tensor):
        feats = [feats]

    feat_shapes = [f.shape for f in feats]
    return feat_shapes


import torch
import numpy as np

class DenseLayer(torch.nn.Module):
    def __init__(self,
                 c_in, # dimensionality of input features
                 c_out, # dimensionality of output features
                 zero_init=False, # initialize weights as zeros; use Xavier uniform init if zero_init=False
                 ):
        super().__init__()
        self.linear = torch.nn.Linear(c_in, c_out)
        if zero_init:
            torch.nn.init.zeros_(self.linear.weight.data)
        else:
            torch.nn.init.uniform_(self.linear.weight.data, -np.sqrt(6 / (c_in + c_out)), np.sqrt(6 / (c_in + c_out)))
        torch.nn.init.zeros_(self.linear.bias.data)
    def forward(self,node_feats):
        node_feats = self.linear(node_feats)
        return node_feats

class DeconvNetPerSpotPatchEffect(torch.nn.Module):
    def __init__(self, gene_num, hidden_dims, n_celltypes, patch_num, patch_emb_dim = 16):
        super(DeconvNetPerSpotPatchEffect, self).__init__()
        self.hidden_dims = hidden_dims
        self.deconv_alpha_layer = DenseLayer(hidden_dims + patch_emb_dim, 1, zero_init=True)

        self.deconv_beta_layer = DenseLayer(hidden_dims, n_celltypes, zero_init=True)
        self.gamma = torch.nn.Parameter(torch.Tensor(patch_num, gene_num).zero_())
        self.patch_emb = torch.nn.Embedding(patch_num, patch_emb_dim)

    def forward(self, z, count_matrix, library_size, basis, patch_index):
        beta, alpha = self.deconv(z, patch_index)
        # Add entropy loss to encourage sparsity
        log_lam = torch.log(torch.matmul(beta, basis) + 1e-6) + alpha + self.gamma[patch_index].unsqueeze(0)
        lam = torch.exp(log_lam)

        decon_loss = -torch.mean(torch.sum(
                count_matrix * (torch.log(library_size + 1e-6) + log_lam) - library_size * lam, dim=1)
            )

        return decon_loss

    def deconv(self, z, patch_index):
        # patch_index: int -> N x 1 tensor
        patch_index = torch.tensor(patch_index).unsqueeze(0).to(z.device)
        patch_index = patch_index.repeat(z.shape[0])
        beta = self.deconv_beta_layer(torch.nn.functional.elu(z))
        beta = torch.nn.functional.softmax(beta, dim=1)
        patch_emb = self.patch_emb(patch_index)
        alpha = self.deconv_alpha_layer(torch.nn.functional.elu(torch.cat([z, patch_emb], dim=1)))
        return beta, alpha
    

class DeconvNetNoPatchEffect(torch.nn.Module):
    def __init__(self, gene_num, hidden_dims, n_celltypes, patch_num, patch_emb_dim = 16):
        super(DeconvNetNoPatchEffect, self).__init__()
        self.hidden_dims = hidden_dims
        self.deconv_alpha_layer = DenseLayer(hidden_dims, 1, zero_init=True)
        self.deconv_beta_layer = DenseLayer(hidden_dims, n_celltypes, zero_init=True)
        self.gamma = torch.nn.Parameter(torch.Tensor(1, gene_num).zero_())

    def forward(self, z, count_matrix, library_size, basis, patch_index):
        beta, alpha = self.deconv(z, patch_index)
        # Add entropy loss to encourage sparsity
        log_lam = torch.log(torch.matmul(beta, basis) + 1e-6) + alpha + self.gamma
        lam = torch.exp(log_lam)

        decon_loss = -torch.mean(torch.sum(
                count_matrix * (torch.log(library_size + 1e-6) + log_lam) - library_size * lam, dim=1)
            )

        return decon_loss

    def deconv(self, z, patch_index):
        beta = self.deconv_beta_layer(torch.nn.functional.elu(z))
        beta = torch.nn.functional.softmax(beta, dim=1)
        alpha = self.deconv_alpha_layer(torch.nn.functional.elu(z))
        return beta, alpha


class NMFNet(torch.nn.Module):
    def __init__(self, gene_num, hidden_dims, n_celltypes, patch_num, patch_emb_dim = 16):
        super(NMFNet, self).__init__()
        self.hidden_dims = hidden_dims
        self.deconv_alpha_layer = DenseLayer(hidden_dims + patch_emb_dim, 1, zero_init=True)

        self.deconv_beta_layer = DenseLayer(hidden_dims, n_celltypes, zero_init=True)
        self.facotors = torch.nn.Parameter(torch.Tensor(n_celltypes, gene_num).
                                           uniform_(-np.sqrt(6/(n_celltypes+gene_num)), np.sqrt(6/(n_celltypes+gene_num))))

        self.gamma = torch.nn.Parameter(torch.Tensor(patch_num, gene_num).zero_())
        self.patch_emb = torch.nn.Embedding(patch_num, patch_emb_dim)

    def get_basis(self):
        basis = torch.nn.functional.softmax(self.facotors, dim=1)
        return basis

    def forward(self, z, count_matrix, library_size, basis, patch_index):
        basis = self.get_basis()
        beta, alpha = self.deconv(z, patch_index)
        # Add entropy loss to encourage sparsity
        log_lam = torch.log(torch.matmul(beta, basis) + 1e-6) + alpha + self.gamma[patch_index].unsqueeze(0)
        lam = torch.exp(log_lam)

        decon_loss = -torch.mean(torch.sum(
                count_matrix * (torch.log(library_size + 1e-6) + log_lam) - library_size * lam, dim=1)
            )

        return decon_loss

    def deconv(self, z, patch_index):
        # patch_index: int -> N x 1 tensor
        patch_index = torch.tensor(patch_index).unsqueeze(0).to(z.device)
        patch_index = patch_index.repeat(z.shape[0])
        beta = self.deconv_beta_layer(torch.nn.functional.elu(z))
        beta = torch.nn.functional.softmax(beta, dim=1)
        patch_emb = self.patch_emb(patch_index)
        alpha = self.deconv_alpha_layer(torch.nn.functional.elu(torch.cat([z, patch_emb], dim=1)))
        return beta, alpha

class CellPredictNet(torch.nn.Module):
    def __init__(self, cond_channels = 128):
        super(CellPredictNet, self).__init__()
        self.cond_channel = cond_channels

        # FLiM
        self.FLiMmul = torch.nn.Sequential(
            torch.nn.Conv2d(self.cond_channel, 32, kernel_size=1, stride=1, padding=0),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(32, 16, kernel_size=1, stride=1, padding=0),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(16, 1, kernel_size=1, stride=1, padding=0),
        )

        self.predict = torch.nn.Conv2d(1, 2, kernel_size=1, stride=1, padding=0)

    def forward(self, nuclei_soft_mask, cond):
        f_mul = self.FLiMmul(cond)
        x = nuclei_soft_mask * f_mul
        x = self.predict(x)
        return x

def deactivate_batchnorm(m):
    if isinstance(m, torch.nn.BatchNorm2d):
        m.reset_parameters()
        m.eval()
        with torch.no_grad():
            m.weight.fill_(1.0)
            m.bias.zero_()


import os
import bin2cell as b2c
import cv2
import scanpy as sc
import pandas as pd
import numpy as np
import scipy
from utils.io import save_list, save_array
from scipy.sparse import coo_matrix

from tqdm import tqdm
import multiprocessing as mp
class SingleCellPreprocessor:
    def __init__(self, adata, celltype_col, save_path, st_gene_list = None, celltype_ref = None, sample_col = None):
        self.adata = adata
        self.celltype_ref = celltype_ref
        self.celltype_col = celltype_col
        self.sample_col = sample_col
        os.makedirs(save_path, exist_ok=True)
        self.save_path = save_path
        self.st_gene_list = st_gene_list if st_gene_list is not None else adata.var_names

    def preprocess(self, hvg_method = None, n_hvg_group = 200, n_hvg = 2000):
        adata_ref = self.adata.copy()
        adata_ref.var_names_make_unique()
        # Remove mt-genes
        adata_ref = adata_ref[:, np.array(~adata_ref.var.index.isna())
                                 & np.array(~adata_ref.var_names.str.startswith("mt-"))
                                 & np.array(~adata_ref.var_names.str.startswith("MT-"))]
        if self.celltype_ref is not None:
            if not isinstance(self.celltype_ref, list):
                raise ValueError("'celltype_ref' must be a list!")
            else:
                adata_ref = adata_ref[[(t in self.celltype_ref) for t in adata_ref.obs[self.celltype_col].values.astype(str)],:]
        else:
            celltype_counts = adata_ref.obs[self.celltype_col].value_counts()
            celltype_ref = list(celltype_counts.index[celltype_counts > 1])
            adata_ref = adata_ref[[(t in celltype_ref) for t in adata_ref.obs[self.celltype_col].values.astype(str)], :]

        # Remove cells and genes with 0 counts
        sc.pp.filter_cells(adata_ref, min_genes=1)
        sc.pp.filter_genes(adata_ref, min_cells=1)

        # Take the subset of genes
        select_index = []
        for g in self.st_gene_list:
            if g in adata_ref.var_names:
                select_index.append(list(adata_ref.var_names).index(g))

        adata_ref = adata_ref[:, select_index]

        if hvg_method == "t-test":
            print("Select highly variable genes using t-test method...")
            # Select hvg
            adata_ref_log = adata_ref.copy()
            sc.pp.log1p(adata_ref_log)
            hvgs = self.select_hvgs(adata_ref_log, n_hvg_group)
            print("%d highly variable genes selected." % len(hvgs))
            adata_ref = adata_ref[:, hvgs]
        elif hvg_method == "seurat_v3":
            print("Select highly variable genes using Seurat v3 method...")
            sc.pp.highly_variable_genes(adata_ref, flavor="seurat_v3", n_top_genes=n_hvg)
            hvg = list(adata_ref.var.loc[adata_ref.var['highly_variable']==1].index)
            adata_ref = adata_ref[:, hvg]

            print("%d highly variable genes selected." % len(adata_ref.var.highly_variable))
        elif hvg_method == "combined":
            print("Select highly variable genes using combined method...")
            # Select hvg
            adata_ref_log = adata_ref.copy()
            sc.pp.log1p(adata_ref_log)
            hvgs_ttest = self.select_hvgs(adata_ref_log, n_hvg_group)

            sc.pp.highly_variable_genes(adata_ref, flavor="seurat_v3", n_top_genes=n_hvg)
            hvgs_seurat = list(adata_ref.var.loc[adata_ref.var['highly_variable']==1].index)
            hvgs = list(set(hvgs_ttest) | set(hvgs_seurat))
            adata_ref = adata_ref[:, hvgs]
            print("%d highly variable genes selected." % len(hvgs))
        elif hvg_method is None:
            print("No highly variable genes selected. All genes are used.")
        else:
            raise ValueError("Unknown hvg_method!")

        print("Calculate basis for deconvolution...")
        sc.pp.filter_cells(adata_ref, min_genes=1)
        sc.pp.normalize_total(adata_ref, target_sum=1)
        celltype_list = list(sorted(set(adata_ref.obs[self.celltype_col].values.astype(str))))
        basis = np.zeros((len(celltype_list), len(adata_ref.var.index)))

        if self.sample_col is not None:
            sample_list = list(sorted(set(adata_ref.obs[self.sample_col].values.astype(str))))
            for i in range(len(celltype_list)):
                c = celltype_list[i]
                tmp_list = []
                for j in range(len(sample_list)):
                    s = sample_list[j]
                    tmp = adata_ref[(adata_ref.obs[self.celltype_col].values.astype(str) == c) &
                            (adata_ref.obs[self.sample_col].values.astype(str) == s), :].X
                    if scipy.sparse.issparse(tmp):
                        tmp = tmp.toarray()
                    if tmp.shape[0] >= 3:
                        tmp_list.append(np.mean(tmp, axis=0).reshape((-1)))
                tmp_mean = np.mean(tmp_list, axis=0)
                if scipy.sparse.issparse(tmp_mean):
                    tmp_mean = tmp_mean.toarray()
                print("%d batches are used for computing the basis vector of cell type <%s>." % (len(tmp_list), c))
                basis[i, :] = tmp_mean
        else:
            for i in range(len(celltype_list)):
                c = celltype_list[i]
                tmp = adata_ref[adata_ref.obs[self.celltype_col].values.astype(str) == c, :].X
                if scipy.sparse.issparse(tmp):
                    tmp = tmp.toarray()
                basis[i, :] = np.mean(tmp, axis=0).reshape((-1))

        # Save
        save_array(basis, self.save_path + "/basis.npy")
        save_list(celltype_list, self.save_path + "/celltype_names.txt")
        save_list(list(adata_ref.var_names), self.save_path + "/filtered_gene_names.txt")

    def select_hvgs(self, adata_ref_log, n_hvg_group):
        sc.tl.rank_genes_groups(adata_ref_log, groupby=self.celltype_col, method="t-test", key_added="ttest", use_raw=False)
        markers_df = pd.DataFrame(adata_ref_log.uns['ttest']['names']).iloc[0:n_hvg_group, :]
        genes = sorted(list(np.unique(markers_df.melt().value.values)))
        return genes


class VisiumHDPreprocessor:
    def __init__(self, bin_dir, source_he_image_path, spaceranger_image_path, save_path):
        self.adata = b2c.read_visium(bin_dir,
                                     source_image_path = source_he_image_path,
                                        spaceranger_image_path = spaceranger_image_path)
        self.adata.var_names_make_unique()
        self.save_path = save_path
        os.makedirs(self.save_path, exist_ok=True)
        # Save st gene list
        save_list(self.adata.var_names, self.save_path + "/st_gene_list.txt")

    def get_nuclei_segmentation(self, mpp = 0.5):
        b2c.scaled_he_image(self.adata, mpp=mpp, save_path = os.path.join(self.save_path, "scaled_he_image.tiff"))

        b2c.stardist(image_path=os.path.join(self.save_path, "scaled_he_image.tiff"),
                     labels_npz_path=os.path.join(self.save_path, "stardist_segmentation.npz"),
                     stardist_model="2D_versatile_he",
                     prob_thresh=0.01
                     )

        b2c.insert_labels(self.adata,
                          labels_npz_path=os.path.join(self.save_path, "stardist_segmentation.npz"),
                          basis="spatial",
                          spatial_key="spatial_cropped_150_buffer",
                          mpp=mpp,
                          labels_key="cell_id",
                          )

    def preprare_sst(self, filtered_gene_list, cell_col = "cell_id", cores = 20):
        adata_st = self.adata.copy()
        adata_st.var_names_make_unique()
        adata_st = adata_st[:, filtered_gene_list]

        map_width = adata_st.obs['array_col'].max() + 1
        map_height = adata_st.obs['array_row'].max() + 1

        adata_st.obs_names_make_unique()
        # Index Mapping
        adata_st.obs["spot_id"] = np.arange(adata_st.shape[0]) + 1
        spot_id_map = coo_matrix((adata_st.obs['spot_id'], (adata_st.obs['array_row'], adata_st.obs['array_col'])),
                                 shape=(map_height, map_width)).toarray()
        np.save(os.path.join(self.save_path, "spot_id_map.npy"), spot_id_map)
        # Save
        adata_st.write(self.save_path + "/filtered_adata_st.h5ad")

        # Turn in to adata
        n_processes = cores
        gene_names_chunks = np.array_split(filtered_gene_list, n_processes)
        processes = []

        # Create temp dir
        temp_dir = os.path.join(self.save_path, "per_gene_map")
        os.makedirs(temp_dir, exist_ok=True)
        for i, gene_chunk in enumerate(gene_names_chunks):
            adata_sub = adata_st[:, gene_chunk].copy()
            p = mp.Process(target=self.process_gene_chunk, args= (adata_sub, temp_dir, map_height, map_width))
            processes.append(p)
            p.start()

        for p in processes:
            p.join()

        # Combine channel-wise
        gene_map = None
        for i, gene in tqdm(enumerate(filtered_gene_list), total=len(filtered_gene_list)):
            map = np.load(os.path.join(temp_dir, gene + ".npy"))
            if gene_map is None:
                gene_map = np.zeros((map.shape[0], map.shape[1], len(filtered_gene_list)), dtype=np.uint8)
                gene_map[:, :, i] = map
            else:
                gene_map[:, :, i] = map

        # Save the combined map
        save_array(gene_map, os.path.join(self.save_path, "gene_map.npy"))

        # Collect segmentation label
        df_seg = adata_st.obs[['array_row', 'array_col', cell_col]]
        seg_map = coo_matrix((df_seg[cell_col], (df_seg['array_row'], df_seg['array_col'])), shape=(map_height, map_width)).toarray()
        save_array(seg_map, os.path.join(self.save_path, "segmentation_mask.npy"))


    def process_gene_chunk(self, adata_sub, temp_dir, map_height, map_width):
        for gene in list(adata_sub.var_names):
            temp_df = adata_sub.obs[['array_row', 'array_col']]
            if scipy.sparse.issparse(adata_sub.X):
                temp_df[gene] = adata_sub[:, gene].X.toarray().flatten()
            else:
                temp_df[gene] = adata_sub[:, gene].X.flatten()
            map = coo_matrix((temp_df[gene], (temp_df['array_row'], temp_df['array_col'])), shape=(map_height, map_width)).toarray()
            save_array(map, os.path.join(temp_dir, gene + ".npy"))
            print(f"Gene {gene} processed.")

class XeniumPreprocessor:
    def __init__(self, transcripts_file, nucleus_boundary_10X, save_path, min_qv = 20):
        if transcripts_file.endswith(".parquet"):
            df = pd.read_parquet(transcripts_file)
        elif transcripts_file.endswith(".csv"):
            df = pd.read_csv(transcripts_file)
        else:
            raise ValueError("Unknown file format!")
        df['feature_name'] = df['feature_name'].apply(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x)
        df = df[(df["qv"] >= min_qv) &
                (~df["feature_name"].str.startswith("NegControlProbe_")) &
                (~df["feature_name"].str.startswith("antisense_")) &
                (~df["feature_name"].str.startswith("NegControlCodeword_")) &
                (~df["feature_name"].str.startswith("BLANK_")) &
                (~df["feature_name"].str.startswith("UnassignedCodeword_"))]
        df.reset_index(inplace=True, drop=True)
        os.makedirs(save_path, exist_ok=True)

        self.nucleus_boundary_10X = nucleus_boundary_10X
        self.save_path = save_path
        df.to_csv(os.path.join(self.save_path, 'transcripts_filtered.csv'))
        self.df = df
        gene_names = df['feature_name'].unique()
        print('%d unique genes' % len(gene_names))
        self.gene_names = sorted(gene_names)
        save_list(gene_names, os.path.join(self.save_path, 'st_gene_list.txt'))

        self.map_width = int(np.ceil(df['x_location'].max())) + 1
        self.map_height = int(np.ceil(df['y_location'].max())) + 1


    # Multi-process of getting nuclei segmentation
    def get_nuclei_segmentation(self, cores = 30):
        nucleus_boundaries = pd.read_parquet(self.nucleus_boundary_10X)
        # All cell_id
        cell_ids = nucleus_boundaries['cell_id'].unique()
        n_processes = cores
        cell_id_chunks = np.array_split(cell_ids, n_processes)
        processes = []
        for i, cell_id_chunk in enumerate(cell_id_chunks):
            begin_mask_id = i * len(cell_id_chunk)
            p = mp.Process(target=self.process_nuclei_segmentation, args=(cell_id_chunk, nucleus_boundaries, begin_mask_id, i))
            processes.append(p)
            p.start()
        for p in processes:
            p.join()
        # Combine
        nuclei_mask_whole = np.zeros((self.map_height, self.map_width))
        for i in range(n_processes):
            current_mask = np.load(os.path.join(self.save_path, f"segmentation_mask_{i}.npy"))
            nuclei_mask_whole[current_mask > 0] = current_mask[current_mask > 0]

        # Save
        save_array(nuclei_mask_whole.astype(np.uint32), os.path.join(self.save_path, "segmentation_mask.npy"))
        # Delete temp files
        for i in range(n_processes):
            os.remove(os.path.join(self.save_path, f"segmentation_mask_{i}.npy"))

    def process_nuclei_segmentation(self, cell_id_chunk, nucleus_boundaries, begin_mask_id, process_id):
        nuclei_mask_whole = np.zeros((self.map_height, self.map_width))
        bar = tqdm(enumerate(cell_id_chunk), total=len(cell_id_chunk))
        for m, id in bar:
            mask_id = m + 1 + begin_mask_id
            single_nuclei_boundary = nucleus_boundaries[nucleus_boundaries['cell_id'] == id]
            single_nuclei_boundary = single_nuclei_boundary.reset_index(drop=True)
            if len(single_nuclei_boundary) > 0:
                # Detect how many nuclei in one cell
                df_nb = single_nuclei_boundary.copy()
                # Find none unique coordinates and its index
                df_nb = df_nb[df_nb.duplicated(subset=['vertex_x', 'vertex_y'], keep=False)]
                poly_num = int(len(df_nb) / 2)

                nuclei_poly_list = []
                for i in range(poly_num):
                    poly_index = (df_nb.index[i * 2], df_nb.index[i * 2 + 1])
                    df_poly = single_nuclei_boundary.loc[poly_index[0]:poly_index[1]]
                    x = df_poly.vertex_x
                    y = df_poly.vertex_y
                    polygons = [list(zip(x, y))]
                    polygons = np.array(polygons, 'int32')
                    nuclei_poly_list.append(polygons)
                cv2.fillPoly(nuclei_mask_whole, nuclei_poly_list, mask_id)
        # Save
        save_array(nuclei_mask_whole.astype(np.uint32), os.path.join(self.save_path, f"segmentation_mask_{process_id}.npy"))

    def preprare_sst(self, filtered_gene_list, cores = 20):
        # Prepare all patches
        n_processes = cores
        gene_names_chunks = np.array_split(filtered_gene_list, n_processes)
        processes = []

        # Create temp dir
        temp_dir = os.path.join(self.save_path, "per_gene_map")
        os.makedirs(temp_dir, exist_ok=True)
        for i, gene_chunk in enumerate(gene_names_chunks):
            df_sub = self.df[self.df['feature_name'].isin(gene_chunk)]
            p = mp.Process(target=self.process_gene_chunk, args=(gene_chunk, df_sub, temp_dir, self.map_width, self.map_height))
            processes.append(p)
            p.start()

        for p in processes:
            p.join()

        # Combine channel-wise
        gene_map = None
        for i, gene in tqdm(enumerate(filtered_gene_list), total=len(filtered_gene_list)):
            map = np.load(os.path.join(temp_dir, gene + ".npy"))
            if gene_map is None:
                gene_map = np.zeros((map.shape[0], map.shape[1], len(filtered_gene_list)), dtype=np.uint8)
                gene_map[:, :, i] = map
            else:
                gene_map[:, :, i] = map

        # Save the combined map
        save_array(gene_map, os.path.join(self.save_path, "gene_map.npy"))

    def process_gene_chunk(self, gene_chunk, df_sub, temp_dir, map_width, map_height):
        for gene in gene_chunk:
            temp_df = df_sub[df_sub['feature_name'] == gene]
            temp_df['x'] = np.round(temp_df['x_location']).astype(int)
            temp_df['y'] = np.round(temp_df['y_location']).astype(int)
            # Add column count
            temp_df['MIDCount'] = 1
            temp_df = temp_df.groupby(['x', 'y']).sum().reset_index()
            map = coo_matrix((temp_df['MIDCount'], (temp_df['y'], temp_df['x'])), shape=(map_height, map_width)).toarray()
            save_array(map, os.path.join(temp_dir, gene + ".npy"))
            print(f"Gene {gene} processed.")


###################################################################################################
# In this file, we define the ExperimentManager class, which is used to manage the experiment
# configuration, logging, and checkpointing. The ExperimentManager class is used in the main
# script to set up the experiment, log the configuration, and save the checkpoints.
#
# The ExperimentManager class has the following methods:
# - get_basic_arg_parser: Returns an argument parser with the basic options for the project.
# - _setup_options: Parses the input arguments and sets the options for the project.
# - _setup_dirs: Sets up the log directory and checkpoint directory for the project.
# - _set_up_logger: Sets up the logger for the project.
# - _setup_seed: Sets the random seed for reproducibility.
# - _setup_torch: Sets up the PyTorch environment.
# - _export_arguments: Saves the experiment configuration to a file.
# - _setup: Sets up the experiment by calling the above methods.
# - get_opt: Returns the experiment options.
# - get_logger: Returns the logger for the project.
# - get_log_dir: Returns the log directory for the project.
# - get_checkpoint_dir: Returns the checkpoint directory for the project.
###################################################################################################

import argparse
import logging
import os
import random
import shutil
import sys
import time
import numpy as np
import torch


class ExperimentManager(object):
    def __init__(self):
        # Options for this project
        self._opt = None
        # Logger for project
        self._logger = None
        # Running result log dir for this project
        self._log_dir = None
        # Checkpoint dir for this project
        self._checkpoint_dir = None

        # Set up the experiment manager
        self._setup()

    def get_basic_arg_parser(self):
        parser = argparse.ArgumentParser()
        # Basic options
        parser.add_argument('--gene_map', type=str, help="gene map tif file, with shape (n, m, k), k is the number of genes")
        parser.add_argument('--nuclei_mask', type=str, help="paired nuclei mask tif file, with shape (n, m)", default=None)
        parser.add_argument('--basis', type=str, help="basis tif file, with shape (c, k), c is the number of cell types, if basis is not provided, use NMF model", default=None)
        parser.add_argument('--celltype_names', type=str, default=None, help="cell names")
        parser.add_argument('--gene_names', type=str, default=None, help="gene names")
        parser.add_argument('--log_dir', type=str, default="./Log", help="run dir for logging, default is ./Log")
        parser.add_argument('--normalized_total', type=int, default=1e3, help="normalized total for gene map")
        parser.add_argument('--min_umi', type=int, default=10, help="minimum umi for a cell")
        parser.add_argument('--min_area', type=int, default=4, help="minimum area for a cell")

        # Model options
        parser.add_argument('--patch_size', type=int, default=400, help="patch size for training and prediction")
        parser.add_argument('--deconv_emb_dim', type=int, default=128, help="embedding dimension for deconvolution")
        parser.add_argument('--no_patch_effect', action='store_true', help="if true, no patch effect in deconvolution")

        # Visualization options
        parser.add_argument('--color_seed', type=int, default=10, help="color seed for visualization")
        parser.add_argument('--point_size', type=float, default=0.5, help="point size for visualization")

        # Training options
        parser.add_argument('--lr_bb', type=float, default=1e-3, help="learning rate for backbone")
        parser.add_argument('--lr_fpn', type=float, default=1e-3, help="learning rate for fpn")
        parser.add_argument('--lr_deconv', type=float, default=1e-2, help="learning rate for deconv")
        parser.add_argument('--lr_decoder', type=float, default=1e-3, help="learning rate for decoder")
        parser.add_argument('--recon_loss_weight', type=float, default=0.1, help="reconstruction loss weight")
        parser.add_argument('--gradient_clip', type=float, default=10, help="gradient clip value")
        parser.add_argument('--num_workers', type=int, default=6, help="number of workers for dataloader")

        parser.add_argument('--save_period', type=int, default=100, help="save period")
        parser.add_argument('--pred_period', type=int, default=100, help="log period")
        parser.add_argument('--save_pixel_feature', action='store_true', help="save pixel feature (If is true, the predicting process will be slow)")


        parser.add_argument('--epochs', type=int, default=400, help="number of epochs for training")
        parser.add_argument('--deconv_warmup_epochs', type=int, default=100, help="warmup epochs for deconv")
        parser.add_argument('--seg_training_epochs', type=int, default=10, help="training epochs for segmentation")

        # Cell segmentation net training
        parser.add_argument("--dilation_kernel_size", default=10, type=int, help="dilation kernel size")
        parser.add_argument("--dilation_iter_num", default=3, type=int, help="dilation iteration number")
        parser.add_argument("--tau", default=5, type=int, help="parameter for creating the soft mask")
        parser.add_argument("--cell_chunk_size", default=320, type=int, help="chunk size for cell segmentation")
        # NMF
        parser.add_argument('--nmf', action='store_true', help="use NMF model")
        parser.add_argument('--factor_num', type=int, default=10, help="factor number for NMF")

        # Environment options
        parser.add_argument('--gpu', type=str, default="0", help="gpu device id")
        parser.add_argument('--seed', type=int, default=0, help="random seed")

        # Load model
        parser.add_argument('--load_model', type=str, default=None, help="load model path")

        return parser

    def _setup_options(self):
        parser = self.get_basic_arg_parser()
        self._opt, _ = parser.parse_known_args()

    def _setup_dirs(self):
        opt = self._opt
        self._log_dir = opt.log_dir
        if os.path.exists(self._log_dir):
            print(f"Log dir exists: {self._log_dir}, please choose an option:")
            op = input("d (delete) / n (new) / q (quit): ")
            if op == 'd':
                shutil.rmtree(opt.log_dir, ignore_errors=True)
                print("Old files deleted.")
            elif op == 'n':
                self._log_dir = opt.log_dir + f"_new_{int(time.time())}"
            else:
                raise OSError("Quit without changes.")
        os.makedirs(self._log_dir, exist_ok=True)
        print(f"Log dir: {self._log_dir}")
        # Checkpoint dir
        self._checkpoint_dir = os.path.join(self._log_dir, "checkpoint")
        os.makedirs(self._checkpoint_dir, exist_ok=True)
        print(f"Checkpoint dir: {self._checkpoint_dir}")

    def _set_up_logger(self, level=logging.DEBUG, name="UCS"):
        """
        Setting self._logger
        """
        self._logger = logging.getLogger(name=name)
        self._logger.propagate = False
        self._logger.setLevel(level)
        # Stdout handler
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(level)
        ch.setFormatter(formatter)
        self._logger.addHandler(ch)
        # Log file handler
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        filename = "log.log"
        path = os.path.join(self._log_dir, filename)
        fh = logging.FileHandler(path, encoding='utf-8')
        fh.setLevel(level)
        fh.setFormatter(formatter)
        self._logger.addHandler(fh)

    def _setup_seed(self):
        random.seed(self._opt.seed)
        np.random.seed(self._opt.seed)
        torch.manual_seed(self._opt.seed)
        torch.cuda.manual_seed_all(self._opt.seed)

    def _setup_torch(self):
        os.environ['CUDA_VISIBLE_DEVICES'] = self._opt.gpu
        torch.backends.cudnn.enabled = True
        torch.backends.cudnn.benchmark = True

    def _export_arguments(self):
        """
        Save self._opt
        """
        opt = self._opt
        self._logger.info(f"Opts: {opt}")
        with open(os.path.join(self._log_dir, 'argv.txt'), 'w') as f:
            print(sys.argv, file=f)

    def _setup(self):
        self._setup_options()
        self._setup_dirs()
        self._set_up_logger()
        self._setup_seed()
        self._setup_torch()
        # self._export_arguments()

    def get_opt(self):
        return self._opt

    def get_logger(self):
        return self._logger

    def get_log_dir(self):
        return self._log_dir

    def get_checkpoint_dir(self):
        return self._checkpoint_dir


if __name__ == "__main__":
    # Test the ExperimentManager
    manager = ExperimentManager()
    opt = manager.get_opt()
    logger = manager.get_logger()
    log_dir = manager.get_log_dir()
    checkpoint_dir = manager.get_checkpoint_dir()
    logger.info(f"Log dir: {log_dir}")
    logger.info(f"Checkpoint dir: {checkpoint_dir}")
    logger.info(f"Options: {opt}")
    logger.info("ExperimentManager test passed.")



from setuptools import Command, find_packages, setup

__lib_name__ = "cellart"
__lib_version__ = "0.0.4"
__description__ = "CellART: a unified framework for extracting single-cell information from high-resolution spatial transcriptomics"
__url__ = "https://github.com/YangLabHKUST/cellart"
__author__ = "Yuheng Chen"
__author_email__ = "ychenlp@connect.ust.hk"

setup(
    name = __lib_name__,
    version = __lib_version__,
    description = __description__,
    url = __url__,
    author = __author__,
    author_email = __author_email__,
    zip_safe = False,
    include_package_data = True,
    # long_description=open("README.md").read(),
    # long_description_content_type="text/markdown", 
    license="MIT",
    license_files=("LICENSE",),
    packages=find_packages(),
    install_requires=[
        "wandb==0.20.0",
        "tensorflow==2.19.0",
        "anndata==0.11.4",
        "bin2cell==0.3.3",
        "distinctipy==1.3.4",
        "einops==0.8.1",
        "matplotlib==3.7.5",
        "numpy==1.26.0",
        "opencv_python==4.10.0.84",
        "opencv_python_headless==4.8.0.76",
        "pandas==2.2.3",
        "scanpy==1.11.2",
        "scipy==1.13",
        "torch==2.0.1",
        "torchvision==0.15.2",
        "tqdm==4.65.0",
    ],
)


# https://github.com/AdeelH/pytorch-fpn/blob/master/fpn/factory.py
from typing import Tuple, Optional
import torch
from torch import nn
import torchvision as tv
from .backbone import ResNetFeatureMapsExtractor
from .layers import Interpolate, SelectOne
from .fpn import FPN, PanopticFPN
from .utils import _get_shapes

def make_spot_fpn_resnet(name: str = 'resnet18',
                         fpn_type: str = 'fpn',
                         out_size: Tuple[int, int] = (224, 224),
                         fpn_channels: int = 256,
                         num_classes: int = 1000,
                         in_channels: int = 500):
    assert in_channels > 0
    assert num_classes > 0

    resnet = tv.models.resnet.__dict__[name](pretrained=False)

    old_conv = resnet.conv1
    old_conv_args = {
        'out_channels': old_conv.out_channels,
        'kernel_size': old_conv.kernel_size,
        'stride': old_conv.stride,
        'padding': old_conv.padding,
        'dilation': old_conv.dilation,
        'groups': old_conv.groups,
        'bias': old_conv.bias
    }
    # just replace the first conv layer
    new_conv = nn.Conv2d(in_channels=in_channels, **old_conv_args)
    resnet.conv1 = new_conv
    backbone = ResNetFeatureMapsExtractor(resnet)

    feat_shapes = _get_shapes(backbone, channels=in_channels, size=out_size)
    if fpn_type == 'fpn':
        fpn = nn.Sequential(
            FPN(feat_shapes,
                hidden_channels=fpn_channels,
                out_channels=num_classes),
            SelectOne(idx=0))
    elif fpn_type == 'panoptic':
        fpn = PanopticFPN(
            feat_shapes,
            hidden_channels=fpn_channels,
            out_channels=num_classes)
    else:
        raise NotImplementedError()

    # yapf: disable
    model = nn.Sequential(
        backbone,
        fpn,
        Interpolate(size=out_size))
    # yapf: enable
    return model